from os import listdir
from os.path import isfile, join, basename, dirname, abspath

directory = dirname(abspath(__file__))

__all__ = [
	f[:-3] # remove the .py
	for f in listdir(directory) # list the ingestors folder
	if isfile(join(directory, f)) and f.endswith('.py') and '__init__' not in f # ensure it ends with .py
]



#!/usr/bin/env python3

from scandeavour.ingestor_base import BaseIngestor
import sys
import os
import xml.etree.ElementTree as ET
import html
import re

class fc:
	red = '\u001b[31m'
	green = '\u001b[32m'
	orange = '\u001b[33m'
	blue = '\u001b[34m'
	end = '\u001b[0m'

class NessusIngestor(BaseIngestor):
	def __init__(self):
		super().__init__(
			name='Nessus Export',
			accepted_files='export.nessus'
		)

	def validate(self, file_path):
		with open(file_path, 'rb') as f:
			# line 1
			line = f.readline()
			if not line.startswith(b'<?xml version'):
				return False
			# line 2
			line = f.readline()
			if not line.startswith(b'<NessusClientData_v2'): 
				return False
			
			self.file_path = file_path
			return True

	def parse(self):
		self.root = None
		try:
			self.root = ET.parse(self.file_path).getroot()
		except Exception as e:
			print(f'{fc.red}[!]{fc.end} Parsing failed with: {e}')
			sys.exit(1)

		nessusreport = self.root.find('Report')

		report_details = self.root.find('.//ReportItem[@pluginID="19506"]/plugin_output')
		# not every report contains this plugin (depends on the scan configuration)
		if report_details is not None:
			report_details = report_details.text
			self.nessusversion = re.search('Nessus version : (?P<version>.+)\n', report_details).group('version')
			self.args = '(Nessus scan name) ' + re.search('Scan name : (?P<name>.+)\n', report_details).group('name')
		else:
			self.nessusversion = ''
			self.args = ''

		self.scan_start = 0
		self.scan_stop = 0

		self.hosts = []
		for xml_host in nessusreport.findall('ReportHost'):

			# host-ip can be ipv6 or ipv4
			tag_ip = xml_host.find('.//tag[@name="host-ip"]')
			ipv4 = ''
			ipv6 = ''
			if tag_ip is not None:
				if '.' in tag_ip.text:
					ipv4 = tag_ip.text
				elif ':' in tag_ip.text:
					ipv6 = tag_ip.text

			# sanity check
			if ipv4 == '' and ipv6 == '':
				print(f'{fc.red}[!]{fc.end} (Nessus Ingestor) host is missing IPv4 and IPv6 - skipping this host')
				continue

			# store hostname
			# currently this ingestor only parses one host-fqdn
			# Nessus also appears to have "host-fqdns" in some cases, which is an array, but I don't have enough test files to implement that properly
			tag_hostname = xml_host.find('.//tag[@name="host-fqdn"]')
			hostnames = []
			if tag_hostname is not None:
				hostnames = [tag_hostname.text]

			# if it exists, add the netbiosname to the hostname
			tag_netbiosname = xml_host.find('.//tag[@name="netbios-name"]')
			if tag_netbiosname is not None:
				hostnames.append(tag_netbiosname.text)

			# store OS name, when reported
			tag_osname = xml_host.find('.//tag[@name="operating-system"]')
			osname = ''
			osconf = 0
			if tag_osname is not None:
				# before adding it, check if a confidence is available
				tag_osconf = xml_host.find('.//tag[@name="operating-system-conf"]')
				if tag_osconf is not None:
					osname = tag_osname.text
					osconf = int(tag_osconf.text) # we hope that this will roughly compare to Nmap (Nmap uses "accuracy" with values between 0 and 100)

			# use host information to update the scan start and stop time
			tag_start = xml_host.find('.//tag[@name="HOST_START_TIMESTAMP"]')
			if tag_start is not None:
				tag_start = int(tag_start.text)
				self.scan_start = tag_start if tag_start < self.scan_start or self.scan_start == 0 else self.scan_start
			tag_stop = xml_host.find('.//tag[@name="HOST_END_TIMESTAMP"]')
			if tag_stop is not None:
				tag_stop = int(tag_stop.text)
				self.scan_stop = tag_stop if tag_stop < self.scan_stop or self.scan_stop == 0 else self.scan_stop

			# list of port objects
			ports = {}
			for report_item in xml_host.findall('ReportItem'):
				xml_port = report_item.attrib
				port_number = int(xml_port.get('port'))
				if port_number == 0:
					continue

				port_protocol = xml_port.get('protocol')
				# Don't rely on Nessus to provide an accurate service name in the svc_name attribute. Usually its something generic like "netbios-ns?"
				# port['svc_name'] = xml_port.get('svc_name', '')

				# Nessus plugins are more interesting though
				script_id = xml_port.get('pluginName')
				plugin_family = xml_port.get('pluginFamily')

				# Nessus generates a lot of output, we focus on these two
				synopsis = report_item.find('synopsis').text # probably mandatory
				plugin_output = report_item.find('plugin_output') # not mandatory

				if (port_id:=f'{port_number}/{port_protocol}') not in ports.keys():
					# store each port with a unique key for the current host
					ports[port_id] = {
						'number': port_number,
						'protocol': port_protocol,
						'scripts': []
					}

				# Only store plugins as "script results" if it is not the port scan plugin itself
				if plugin_family != 'Port scanners':
					output = html.unescape(plugin_output.text) if plugin_output is not None else synopsis
					output = output[:2000]+'... [Content clipped for brevity. See original Nessus result for full plugin output.]' if len(output) > 2000 else output
					ports[port_id]['scripts'].append({
						'id': script_id,
						'output': output
					})

			self.hosts.append({
				'ipv4': ipv4,
				'ipv6': ipv6,
				'ports': [p for p in ports.values()],
				'hostnames': hostnames,
				'osname': osname,
				'osconf': osconf
			})


	def getDatabaseInterface(self):
		scan = {
			'tool': 'Nessus',
			'args': self.args,
			'start': self.scan_start,
			'stop': self.scan_stop,
			'version': self.nessusversion,
			'hosts_up': len(self.hosts),
			'hosts_scanned': len(self.hosts), # not implemented, since there is no easy way of getting this information
			'hosts': [{
				'ipv4': host['ipv4'],
				'ipv6': host['ipv6'],
				'mac': '', # not implemented
				'names': host['hostnames'],
				'reason': '', # not implemented
				'os_name': host['osname'],
				'os_accuracy': host['osconf'],
				'os_family': '', # not implemented
				'os_vendor': '', # not implemented
				'ports': [{
					'port': port['number'],
					'protocol': port['protocol'],
					'svc_name': '', # not implemented
					'svc_info': '', # not implemented
					'svc_ssl': '', # not implemented
					'scripts': port['scripts']
				} for port in host["ports"]],
			} for host in self.hosts]
		}
		return scan


	def __str__(self):
		str = ''
		str += f'CMD: {self.args}\n'
		str += f'Scan started: {self.scan_start}\n'
		str += f'Scan stopped: {self.scan_stop}\n'
		str += f'Nessus version: {self.nessusversion}\n'
		str += f'Scanned hosts ({len(self.hosts)}):\n'
		for host in self.hosts:
			str += f'====\n'
			str += f'\tIPv4: {host["ipv4"]}\n'
			str += f'\tIPv6: {host["ipv6"]}\n'
#			str += f'\tMAC: {host["mac"]}'
			str += f'\tNames: '+ ', '.join(host['hostnames']) + '\n'
#			str += f'\tUp-Reason: {host["up_reason"]}\n'
			str += f'\tOS name: {host["osname"]}\n'
			str += f'\tOS accuracy: {host["osconf"]}\n'
#			str += f'\tOS family: {host["osfamily"]}\n'
#			str += f'\tOS vendor: {host["vendor"]}\n'
			str += f'\tPorts:\n'
			for port in host['ports']:
				str += f'\t\t{port["number"]:5}/{port["protocol"]} \n'
#				str += f'{port["svc_name"]}'
#				str += f'{port["svc_product"]} '
#				str += f'{port["svc_version"]} '
#				str += f'{port["svc_extrainfo"]} '
#				str += f'{port["svc_tunnel"]} '
#				str += f'{port["svc_proto"]}\n'
				for script in port['scripts']:
					str += f'{script["id"]}: {script["output"]}\n'
		return str

# Standalone version
# Requires a copy of ingestor_base.py in the same directory
if __name__ == '__main__':
	if len(sys.argv) != 2:
		print(f'Usage: {sys.argv[0]} nessus_export.xml')
		sys.exit(1)

	inp = sys.argv[1]
	if not os.path.isfile(inp):
		print(f'{fc.red}[!]{fc.end} {inp} does not exist')
		sys.exit(1)

	ing = NessusIngestor()
	if not ing.validate(inp):
		print(f'{fc.red}[!]{fc.end} Not a valid Nessus file')
		sys.exit(1)
	ing.parse()
	print(ing)



from dash import html, dcc, callback, Input, Output, State, register_page
import dash_bootstrap_components as dbc
from scandeavour.components.customToast import CustomToast

register_page(__name__)

def layout(**kwargs):
	return html.Div([
		html.H1("Are you lost?"),
		dcc.Interval(
			id='toast-404',
			n_intervals=0,
			max_intervals=0,
			interval=1
		)
	], style={
		'display': 'flex',
		'alignItems': 'center',
		'justifyContent': 'center',
		'height': '100vh'
	})


@callback(Output('div-toaster', 'children', allow_duplicate=True),
		Input('toast-404', 'n_intervals'),
		State('div-toaster', 'children'),
		prevent_initial_call=True)
def _cb_404toast(_, toasts):
	toasts.append(CustomToast(
		[
			"You tried to access a page that does not exist."
		],
		headerText="404 Not Found",
		level='error'
	))
	return toasts



__version__ = "1.3.1"

from dash import Dash, html, get_asset_url, page_container, DiskcacheManager, clientside_callback, Input, Output, State
import dash_bootstrap_components as dbc
from scandeavour.utils import initDB, TagRibbons
import diskcache # for background callbacks: https://dash.plotly.com/background-callbacks
from flask import Flask, request
import uuid
import tempfile
from os import path, environ
import dash_cytoscape as cyto
import argparse

cyto.load_extra_layouts()
server = Flask(__name__)

@server.route('/upload-file', methods=['POST'])
def upload_file_api():
	try:
		print(f'[+] Uploading file...')
		name = uuid.uuid4().hex
		with open(path.join(tempfile.gettempdir(),name), 'wb') as f:
			chunk_size = 4096
			while True:
				# Use a stream reader, otherwise the app will crash for files with ~400 MB
				chunk = request.stream.read(chunk_size)
				print(f'getting chunk {len(chunk)}')
				if len(chunk) == 0:
					break
				f.write(chunk)
			print(f'[+] File uploaded successfully ({name})')
		return name, 200
	except Exception as e:
		print(f'[!] Failed to process API request: {e}')
		return '', 500


def DashApp(server):

	# Background manager required for the file upload task
	cache = diskcache.Cache("./cache")
	background_callback_manager = DiskcacheManager(cache)

	app = Dash(
		__name__,
		server=server,
		serve_locally=True,
		external_stylesheets=[
			# To make this app load offline, we download the desired theme from bootswatch
			# in this case: DARKLY
			# and then inside the bootstrapmin.css file we must remove the google font import at the
			# very beginning because otherwise this app wil not load offline.
			'/assets/bootstrap.min.css', # https://cdn.jsdelivr.net/npm/bootswatch@5.3.3/dist/darkly/bootstrap.min.css
		],
		use_pages=True,
		suppress_callback_exceptions=True, # required for dynamic components
		background_callback_manager = background_callback_manager, # use for background callbacks
	)

	sidebar = html.Div(
		[
			html.Img(src=get_asset_url('icons/hexagon.svg'), className="nav-logo"),
			html.Hr(),
			dbc.Nav(
				[
					dbc.NavLink(
						[
							html.Img(src=get_asset_url('icons/import.svg')),
							html.Span("Import Scan")
						],
						href="/",
						active="exact"
					),
					dbc.NavLink(
						[
							html.Img(src=get_asset_url('icons/graph.svg')),
							html.Span("View Graph")
						],
						href="/graph",
						active="exact"
					),
					dbc.NavLink(
						[
							html.Img(src=get_asset_url('icons/stats.svg')),
							html.Span("View Data")
						],
						href="/data",
						active="exact"
					),
					dbc.NavLink(
						[
							html.Img(src=get_asset_url('icons/pie-chart.svg')),
							html.Span("View Statistics")
						],
						href="/stats",
						active="exact"
					),
				],
				vertical=True,
				pills=True
			),
			html.Span(f'v{__version__}', className='nav-version')

		],
		id='div-sidebar',
		className="sidebar"
	)

	content = html.Div(
		[
			page_container
		],
		id='div-content',
		className='content',
	)

	toasts = dbc.Container(
		[],
		id='div-toaster',
		class_name='toaster'
	)

	app.layout = html.Div([
		sidebar,
		content,
		toasts,
	])

	return app

def main():
	parser = argparse.ArgumentParser(
		description='Version: ' + __version__,
	)
	parser.add_argument('projectfile', action='store', help='SQLite database project file (will be created if it doesn\'t exist yet)')
	parser.add_argument('-p', '--port', required=False, default=8050, action='store', help='port to host the dashboard on')
	parser.add_argument('-d', '--debug', required=False, action='store_true', help='enable debugging')
	parser.add_argument('-v', '--version', required=False, action='store_true', help='print version')

	args = parser.parse_args()

	if args.version == True:
		print(f'v{__version__}')
		exit(0)

	environ['SQLITE_PROJECT_FILE'] = args.projectfile

	print(f'[+] Version: {__version__}')
	initDB()
	app = DashApp(server)

	# Client side callbacks
	# Placing the clientside_callbacks apparently has to happen before the run_server is executed
	# It wouldn't work in the pages.

        # Delete the graph node history when deleting scan results
	clientside_callback(
		"""
		function(n_clicks){
			if (n_clicks){
				sessionStorage.removeItem('node-nav-state');
				sessionStorage.removeItem('node-nav-state-timestamp');
			}
			return dash_clientside.no_update
		}
		""",
		Output('btn-confirm-delete-scans', 'n_clicks'), # dummy output
		Input('btn-confirm-delete-scans', 'n_clicks'),
	)
	# Recenter the view of the graph without redrawing the layout
	clientside_callback(
		"""
		function(n_clicks){
			cy.fit(padding=20);
			return dash_clientside.no_update;
		}
		""",
		Output('btn-fit-graph-view', 'n_clicks'), # dummy output
		Input('btn-fit-graph-view', 'n_clicks'), # center when clicked
	)
	# Center the graph around a selected node
	clientside_callback(
		"""
		function(n_clicks, unused){
			cy.centre(cy.elements(':selected'));
			return dash_clientside.no_update;
		}
		""",
		Output('btn-fit-graph-view', 'n_clicks', allow_duplicate=True), # dummy output
		Input('btn-center-selected-node', 'n_clicks'), # This input centers a node on button click
		prevent_initial_call=True
	)
	# Regain focus of subnet filter after clicking on it while node info is open
	# First, we need to close the node info (and set the click-intent), then we update the focus
	clientside_callback(
		"""
		function(n_clicks) {
			if (n_clicks>0) {
				return [false, {'id': 'subnet-filter-input'}];
			}
			return [dash_clientside.no_update, {'id': ''}];
		}
		""",
		Output('node-info-canvas', 'is_open', allow_duplicate=True),
		Output('click-intent', 'data', allow_duplicate=True),
		Input('filter-input-wrapper', 'n_clicks'),
		prevent_initial_call=True,
	)
	clientside_callback(
		"""
		function(canvasOpen, clickIntent) {
			if (canvasOpen===false && clickIntent.id.length > 0) {
				setTimeout(function(){
					document.getElementById('subnet-filter-input').focus();
				}, 400); /* This delay ensures, that the offcanvas had time to close */
			}
			return {'id':''};
		}
		""",
		Output('click-intent', 'data', allow_duplicate=True),
		Input('node-info-canvas', 'is_open'),
		State('click-intent', 'data'),
		prevent_initial_call=True
	)
	# Recenter the view of the graph after loading cola layout
	clientside_callback(
		"""
		function(trigger){
			setTimeout(function(){
				cy.fit(padding=20);
			},500); /* Set a timeout to let the layout run a bit*/
			return dash_clientside.no_update;
		}
		""",
		Output('render-trigger', 'data', allow_duplicate=True), # dummy output
		Input('render-trigger', 'data'), # center when graph was redrawn
		prevent_initial_call=True,
	)
	# Overlay button callbacks to update the settings
	# We do this on the clientside because we need access to the datastore of another page (which may not exist yet)
	# Splitting this up into client and server side callbacks is not a good idea
	# because we would update the graph settings from multiple locations and we don't want that. We want to update
	# the graph settings only in one place so the settings are up to date on page rendering and the graph stays synced with
	# the settings.
	clientside_callback(
		"""
		function(n_scans, n_ports, n_hosts, n_datafilter, graph_settings) {
			const triggered_id = dash_clientside.callback_context.triggered_id;

			/* Toggle scan view */
			if (triggered_id === 'btn-toggle-scans' && n_scans > 0) {
				graph_settings['scans-visible'] = !graph_settings['scans-visible'];
			}
			let scans_btn_img_src = '/assets/icons/graphpage/terminal-off.svg';
			let toggle_scans_label = 'Show scans';
			if (graph_settings['scans-visible']) {
				scans_btn_img_src = '/assets/icons/graphpage/terminal-on.svg';
				toggle_scans_label = 'Hide scans';
			}

			/* Toggle port view */
			if (triggered_id === 'btn-toggle-ports' && n_ports > 0) {
				graph_settings['ports-visible'] = !graph_settings['ports-visible'];
			}
			let ports_btn_img_src = '/assets/icons/graphpage/box-off.svg';
			let toggle_ports_label = 'Show ports';
			if (graph_settings['ports-visible']) {
				ports_btn_img_src = '/assets/icons/graphpage/box-on.svg';
				toggle_ports_label = 'Hide ports';
			}

			/* Toggle hosts with no ports */
			if (triggered_id === 'btn-toggle-hosts' && n_hosts > 0) {
				graph_settings['hostsnp-visible'] = !graph_settings['hostsnp-visible'];
			}
			let hosts_btn_img_src = '/assets/icons/graphpage/host-off.svg';
			let toggle_hosts_label = 'Show hosts with no open ports';
			if (graph_settings['hostsnp-visible']) {
				hosts_btn_img_src = '/assets/icons/graphpage/host-on.svg';
				toggle_hosts_label = 'Hide hosts with no open ports';
			}

			/* Toggle datafilter linking */
			if (triggered_id === 'btn-toggle-datafilter' && n_datafilter > 0) {
				graph_settings['datafilter-linked'] = !graph_settings['datafilter-linked'];
			}
			let datafilter_btn_img_src = '/assets/icons/graphpage/datafilter-off.svg';
			let toggle_datafilter_label = 'Apply filters from the data view';
			graph_settings['filters']['datatable'] = [];
			if (graph_settings['datafilter-linked']) {

				datafilter_btn_img_src = '/assets/icons/graphpage/datafilter-on.svg';
				toggle_datafilter_label = 'Ignore filters from the data view';

				/* Get the datatable settings (they may not exist yet) */
				const datatable_settings = sessionStorage.getItem('datatable-settings');

				graph_settings['filters']['datatable'] = datatable_settings ? (JSON.parse(datatable_settings))['filters'] : [];
			}

			return [
				graph_settings,
				scans_btn_img_src,
				ports_btn_img_src,
				hosts_btn_img_src,
				datafilter_btn_img_src,
				toggle_scans_label,
				toggle_ports_label,
				toggle_hosts_label,
				toggle_datafilter_label,
			];
		}
		""",
		Output('graph-settings', 'data'),
		Output('toggle-scans-img', 'src'),
		Output('toggle-ports-img', 'src'),
		Output('toggle-hosts-img', 'src'),
		Output('toggle-datafilter-img', 'src'),
		Output('btn-toggle-scans-label', 'children'),
		Output('btn-toggle-ports-label', 'children'),
		Output('btn-toggle-hosts-label', 'children'),
		Output('btn-toggle-datafilter-label', 'children'),
		Input('btn-toggle-scans', 'n_clicks'),
		Input('btn-toggle-ports', 'n_clicks'),
		Input('btn-toggle-hosts', 'n_clicks'),
		Input('btn-toggle-datafilter', 'n_clicks'),
		State('graph-settings', 'data'),
	)

	# Toggle the tags of the host-detail card in the datapage
	# We don't need to sync toggle switches on page load, as they persist their property via dbc persistence (session)
	# We just need to update the styles of the current host-detail cards and store the change (so the host details can use the setting when rendering)
	clientside_callback(
		"""
		function(toggle_tags, host_detail_settings) {
			host_detail_settings['show-tags'] = toggle_tags;

			const ribbons = document.getElementsByClassName('hd-ribbon');
			let ix=0;
			while (ix < ribbons.length) {
				ribbons[ix].style.display = toggle_tags ? 'block' : 'none';
				ix++;
			}

			return host_detail_settings;
		}
		""",
		Output('host-detail-settings', 'data'),
		Input('btn-toggle-tags', 'value'),
		State('host-detail-settings', 'data'),
	)

	# Update tags on the clientside so that a refresh is not needed when updating tags
	clientside_callback(
		"""
		function(update) {
			const tag_ribbons_map="""+str(TagRibbons.map)+""";
			const targets = new Set(update['targets']);
			const tag = update['tag'];
			const tag_type = update['tag_type'];

			/* Update details */
			const ribbons = document.getElementsByClassName('hd-ribbon');
			let ix=0;
			while (ix < ribbons.length) {
				/* the id of a ribbon is an object like {"index": "btn..", "type": "<hid>"} */
				if (targets.has(JSON.parse(ribbons[ix].id)['type'])) {
					ribbons[ix].style.background = tag_ribbons_map[tag_type]['css-color'];
					ribbons[ix].innerText = tag;
				}
				ix++;
			}

			/* Update AG grid */
			const vdtApi = dash_ag_grid.getApi('view-data-table')
			const table_rows = vdtApi.rowModel.rowsToDisplay;
			ix = 0;
			while (ix < table_rows.length) {
				if (targets.has(table_rows[ix].data.hid.toString())) {
					vdtApi.rowModel.rowsToDisplay[ix].data.tag = tag === Object.keys(tag_ribbons_map)[0] ? '' : tag;
				}
				ix++;
			}
			/* activate changes */
			vdtApi.refreshCells();

			return {};
		}
		""",
		Output('trigger-tag-update', 'data', allow_duplicate=True),
		Input('trigger-tag-update', 'data'),
		prevent_initial_call=True
	)

	# DO NOT EXPOSE THIS APPLICATION TO ANYTHING OTHER THAN LOCALHOST
	app.run(
		host='127.0.0.1',
		port=args.port,
		debug=args.debug,
		use_reloader=False, # using this option can lead to unexpected behaviour (calls being executed twice etc.)
		dev_tools_silence_routes_logging=(not args.debug)
	)

if __name__ == "__main__":
	main()



#!/usr/bin/env python3

# Import the plaintext results of masscan in the format of:
#
# Discovered open port 22/tcp on 127.0.0.1
# Discovered open port 80/tcp on 127.0.0.1

from scandeavour.ingestor_base import BaseIngestor
import sys
import os
import xml.etree.ElementTree as ET

class fc:
	red = '\u001b[31m'
	green = '\u001b[32m'
	orange = '\u001b[33m'
	blue = '\u001b[34m'
	end = '\u001b[0m'

class MasscanPlainIngestor(BaseIngestor):
	def __init__(self):
		super().__init__(
			name = 'MasscanPlain',
			accepted_files = 'masscan.log'
		)

	def validate(self, file_path):
		# masscan plain output contains always the same line
		with open(file_path, 'rb') as f:
			line = f.readline()
			if line.startswith(b'Discovered open port '):
				self.file_path = file_path
				return True
		return False

	def parse(self):

		self.hosts = {}
		with open(self.file_path, 'rb') as f:
			for line in f:
				# parse output from
				# https://github.com/robertdavidgraham/masscan/blob/a31feaf5c943fc517752e23423ea130a92f0d473/src/output.c#L757

				result_line = line.decode()
				if not result_line.startswith('Discovered open port '):
					# ignore closed ports
					continue

				port = result_line.split('open port ')[1].split(' on ')[0]
				port_num = int(port.split('/')[0])
				port_prot = port.split('/')[1]

				if port_prot not in ['tcp', 'udp']:
					# let's ignore arp and icmp results for now
					# it could be added on request though
					continue

				host = result_line.split(' on ')[1]

				if host not in self.hosts.keys():
					self.hosts[host] = []

				self.hosts[host].append({
					'num': port_num,
					'prot': port_prot
				})

	def getDatabaseInterface(self):
		scan = {
			'tool': 'Masscan',
			'args': '', # not implemented by masscan
			'start': 0, # not included in default stdout
			'stop': 0, # not included in default stdout
			'version': '', # not implemented by masscan
			'hosts_up': len(self.hosts),
			'hosts_scanned': len(self.hosts), # not implemented by masscan
			'hosts': [{
				'ipv4': host if '.' in host else '',
				'ipv6': host if ':' in host else '',
				'mac': '', # not yet implemented here
				'names': [], # not implemented by masscan
				'reason': '', # not implemented by masscan
				'os_name': '', # not implemented by masscan
				'os_accuracy': 0, # not implemeneted by masscan
				'os_family': '', # not implemented by masscan
				'os_vendor': '', # not implemented by masscan
				'ports': [{
					'port': port['num'],
					'protocol': port['prot'],
					'svc_name': '', # not implemented
					'svc_info': '', # not implemented
					'svc_ssl': '', # not implemented
					'scripts': [] # not implemented
				} for port in self.hosts[host]],
			} for host in self.hosts]
		}
		return scan


	def __str__(self):
		str = ''
#		str += f'CMD: {self.cmdline}\n'
#		str += f'Scan started: {self.scanstart}\n'
#		str += f'Scan stopped: {self.scanstop}\n'
#		str += f'Masscan version: {self.msversion}\n'
		str += f'Identified hosts: {len(self.hosts)}\n'
		for host in self.hosts:
			str += f'====\n'
			str += f'\tAddr: {host}\n'
#			str += f'\tNames: '+ ', '.join(host['hostnames']) + '\n'
#			str += f'\tUp-Reason: {host["up_reason"]}\n'
#			str += f'\tOS name: {host["osname"]}\n'
#			str += f'\tOS accuracy: {host["accuracy"]}\n'
#			str += f'\tOS family: {host["osfamily"]}\n'
#			str += f'\tOS vendor: {host["vendor"]}\n'
			str += f'\tPorts:\n'
			for port in host[host]:
				str += f'\t\t{port["num"]:5}/{port["prot"]}\n'
#				str += f'{port["svc_name"]} ({port["svc_conf"]}|10) '
#				str += f'{port["svc_product"]} '
#				str += f'{port["svc_version"]} '
#				str += f'{port["svc_extrainfo"]} '
#				str += f'{port["svc_tunnel"]} '
#				str += f'{port["svc_proto"]}\n'
#				for script in port['scripts']:
#					str += f'{script["id"]}: {script["output"]}\n'
		return str

# Standalone version
# Requires a copy of ingestor_base.py in the same directory
if __name__ == '__main__':
	if len(sys.argv) != 2:
		print(f'Usage: {sys.argv[0]} masscan_scan_output.log')
		sys.exit(1)

	inp = sys.argv[1]
	if not os.path.isfile(inp):
		print(f'{fc.red}[!]{fc.end} {inp} does not exist')
		sys.exit(1)

	ing = MasscanIngestor()
	if not ing.validate(inp):
		print(f'{fc.red}[!]{fc.end} Not a valid Masscan file')
		sys.exit(1)
	ing.parse()
	print(ing)



# This is just the base class with pseudo code for demonstration in the validate/parse logic
class BaseIngestor:

	def __init__(self, name, accepted_files):
		self.name = name
		self.accepted_files = accepted_files

	def getName(self):
		return self.name

	def getAcceptedFiles(self):
		return self.accepted_files

	def validate(self, file_path):
		# Run checks to ensure that this ingestor is the right one for this file
		# Do not use .read() as this does not perform well on large files
		# Optimally you read the file line by line for whatever validation you need
		linecount = 0
		with open(file_path, 'rb') as f:
			for line in f:
				linecount += 1
				if b'magic' in line:
					# Save the file path 
					self.file_path = file_path
					# Return True to signal that the ingestor will parse this file
					return True
				if linecount > 20:
					break
		# Return False if this ingestor is not suitable
		return False

	def parse(self):
		# Parse file contents and store relevant information
		with open(self.file_path, 'rb') as f:
			for line in f:
				self.host = line.decode('utf8')
				self.port = 0

	def getDatabaseInterface(self):
		# Return this mandatory scan interface
		scan = {
			'tool': '', 				# * [str] tool name
			'args': '', 				#   [str] scan arguments
			'start': 0, 				#   [int] timestamp scan start
			'stop': 0, 				#   [int] timestamp scan stop
			'version': '',				#   [str] tool version
			'hosts_up': 0,				# * [int] hosts identified as up
			'hosts_scanned': 0,			# * [int] hosts scanned
			'hosts': [{				#   <arr>[obj] hosts identified as up
				'ipv4': '',			# X [str] ipv4
				'ipv6': '',			# X [str] ipv6
				'mac': '',			# X [str] mac
				'names': [],			#   <arr>[str] list of hostnames
				'reason': '',			#   [str] reason why this host is marked alive
				'os_name': '',			#   [str] OS name (e.g. Windows Server 2016)
				'os_accuracy': 0, 		#   [int] OS accuracy
				'os_family': '',		#   [str] OS family (e.g. Windows)
				'os_vendor': '',		#   [str] OS vendor (e.g. Microsoft)
				'ports': [{			#   <arr>[obj]
					'port': 0,		# * [int] port number
					'protocol': '',		# * [str] port protocol (e.g. tcp)
					'svc_name': '', 	#   [str] service name
					'svc_info': '', 	#   [str] service details
					'svc_ssl': '',		#   [str] service tunnel
					'scripts': [{		#   <arr>[obj] scripts for port
						'id': '',	# * [str] script name
						'output': '',	# * [str] script output
					}],			#
				}],				# * = must exist (no * = can be set to empty value)
			}]					# X = at least one must exist
		}
		return scan



from dash import html, dcc, callback, Input, Output, State, register_page, get_asset_url, set_props
from dash.exceptions import PreventUpdate
import dash_ag_grid as dag
import dash_bootstrap_components as dbc
from scandeavour.components.customToast import CustomToast
from scandeavour.ingestors import *
from scandeavour.ingestor_base import BaseIngestor
import sys
import inspect
from scandeavour.utils import getDB, IPtoNum, NumToIP, getHostLabel
from base64 import b64decode
from os import path, remove, stat
import tempfile
import time

register_page(__name__, path='/')

def layout(**kwargs):

	available_ingestors = []
	ingestor_modules = inspect.getmembers(sys.modules['scandeavour.ingestors'], inspect.ismodule)
	for ingestor_module in ingestor_modules:
		mod_name, mod = ingestor_module
		ingestor_classes = inspect.getmembers(mod, inspect.isclass)
		for ingestor in ingestor_classes:
			ingestor_name, ingestor_class = ingestor
			if issubclass(ingestor_class, BaseIngestor) and ingestor_name != 'BaseIngestor':
				available_ingestors.append(ingestor_class())

	accepted_files = ', '.join([ing.getAcceptedFiles() for ing in available_ingestors])

	return 	html.Div([
			# Store value that is cleared when browser tab closes
			dcc.Store(id='upload-data-change', data=False, storage_type='session'),
			dcc.Upload(
				id='upload-scan',
				children=html.Div([
						html.Img(src=get_asset_url('icons/upload-file.svg'), id='import-scan-file-image'), # updated during upload
						html.H4('Import scan', id='import-scan-h4label'), # updated during upload
						html.Span('Select a file or use drag and drop.'),
						html.Span(f'Accepted file types: {accepted_files}'),
					],
					className='upload-content'
				),
				className='upload-scan',
				style={
					'margin': '1rem',
					'borderWidth': '1px',
					'borderStyle': 'dashed',
					'borderRadius': '1rem',
					'textAlign': 'center',
					'padding': '1rem'
			        },
				style_active={
					'backgroundColor': 'var(--bs-secondary)',
				},
				multiple=True,
				disabled=False
			),
			html.Div([
				dbc.Progress(
					id="upload-progress",
					style={'visibility': 'hidden'},
					animated=True,
					striped=True,
					color='success',
					class_name='label-progress-bar',
					label=''
				)
			]),
			html.Div([
				html.Img(src=get_asset_url('icons/headings/list.svg')),
				html.H3('Imported files')
			], className='medium-heading'),
			dcc.Loading([
				dag.AgGrid(
					id='input-files-table',
					rowData=[{'type': 'loading', 'filesize': 0}],
					columnDefs=[
						{'headerName':'Type', 'field':'type', 'maxWidth':150, 'filter': 'agTextColumnFilter'},
						{'headerName':'Filename', 'field':'filename', 'filter': 'agTextColumnFilter'},
						{'headerName':'Hosts scanned', 'field':'targetsScanned', 'maxWidth': 150, 'filter': 'agNumberColumnFilter'},
						{'headerName':'Up', 'field':'targetsUp', 'maxWidth': 100, 'filter': 'agNumberColumnFilter'},
						{'headerName':'Ingest date', 'field':'ingestDate', 'maxWidth': 200},
						{'headerName':'Filesize', 'field':'filesize', 'maxWidth': 150, 'filter': 'agNumberColumnFilter', 'valueFormatter':{'function':'FileSize(params.value)'}}
					],
					defaultColDef={'useValueFormatterForExport': 'False'},
					className='ag-theme-alpine-dark input-file-table-container',
					dashGridOptions={'domLayout': 'autoHeight'},
					columnSize='responsiveSizeToFit',
					style= {
						'height': None,
						'marginBottom': '0.5rem'
					},
				)],
				overlay_style={'visibility': 'visible', 'filter': 'blur(2px)'},
				type='circle',
				delay_hide=200,
				color='#00ffe0'
			),
			dbc.Button('Delete all imported scans', id='btn-delete-scans', disabled=False, outline=True, color='danger', class_name='me-1 scan-del-btn'),
			dbc.Modal(
				[
					dbc.ModalHeader(dbc.ModalTitle('Delete scans')),
					dbc.ModalBody('This will delete all scans stored in the database. Are you sure?'),
					dbc.ModalFooter(
						dbc.Button(
							'Delete',
							id='btn-confirm-delete-scans',
							outline=True,
							color='danger',
							n_clicks=0
						)
					),
				],
				id='modal-confirm-delete-scans',
				centered=True,
				is_open=False
			),
		],
		className='upload-container')


@callback(	Output('upload-data-change', 'data'),
		Output('div-toaster', 'children', allow_duplicate=True),
		Input('upload-scan', 'contents'),
		State('upload-scan', 'filename'),
		State('div-toaster', 'children'),
		background=True, # Required, so we can use "progress" here, for realtime updates
		running=[
			# view progress bar while uploading
			(
				Output('upload-progress', 'style'),
				{"visibility": "visible"},
				{"visibility": "hidden"}
			),
			# prevent database deletion
			(
				Output('btn-delete-scans', 'disabled'),
				True,
				False
			),
			# prevent new upload
			(
				Output('upload-scan', 'disabled'),
				True,
				False
			),
			# change label
			(
				Output('import-scan-h4label', 'children'),
				'Working on it...',
				'Import scan',
			),
			# change image
			(
				Output('import-scan-file-image','src'),
				get_asset_url('icons/coffee.svg'),
				get_asset_url('icons/upload-file.svg')
			)
		],
		progress=[
			Output('upload-progress', 'value'),
			Output('upload-progress', 'label')
		],
		prevent_initial_call=True,
	) # see https://dash.plotly.com/background-callbacks for explanation of set_progress
def _cb_fileUpload(set_progress, file_contents, file_names, toasts):
	# This gets called on upload but we patched file_contents to be empty, to not crash the browser.
	# By now the file contents are uploaded already.
	# See assets/patchFileReader.js.
	if not file_names:
		raise PreventUpdate

	print(f'[+] Parsing uploaded file(s):')
	for ix in range(len(file_names)):
		print(f'\t- ({ix+1}) {file_names[ix]}')

	# Load ingestors
	available_ingestors = []
	try:
		ingestor_modules = inspect.getmembers(sys.modules['scandeavour.ingestors'], inspect.ismodule)
		for ingestor_module in ingestor_modules:
			mod_name, mod = ingestor_module
			ingestor_classes = inspect.getmembers(mod, inspect.isclass)
			for ingestor in ingestor_classes:
				ingestor_name, ingestor_class = ingestor
				if issubclass(ingestor_class, BaseIngestor) and ingestor_name != 'BaseIngestor':
					available_ingestors.append(ingestor_class)
	except Exception as e:
		toasts.append(CustomToast(
			[ f'Error: {e}' ],
			headerText=f'Failed to load ingestors',
			level='error',
			duration=10000,
		))
		file_contents = None

	if len(available_ingestors) < 1:
		toasts.append(CustomToast(
			[ f'At least one ingestor is required in the "ingestors" folder.' ],
			headerText=f'No ingestors found',
			level='error',
			duration=10000,
		))
		file_contents = None

	# For the progress calculations
	# we start at 40/100 when we are here (40% is set by the assets/patchFileReader.js)
	# so we got 60 left to fill.
	parse_bpct = 40									# start of progress bar (base percentage)
	parse_ftotal = len(file_names)					# files to process
	parse_fcnt = 0									# files processed
	parse_fpct = int((100-parse_bpct)/parse_ftotal)	# percentages available per file (regardless of their size)
	

	# Files have been uploaded to a tmp dir (paths are in file_contents)
	# real file names are in file_names and
	# ingestors have been loaded (at least one)
	if file_contents is not None:
		for tmp_name, name in zip(file_contents, file_names):
			# Increase counter immediately, because it may fail anytime and then we continue with the next one
			parse_fcnt += 1

			# Set progress to base percentage + start of nth file
			set_progress((
				parse_bpct+parse_fpct*(parse_fcnt-1),
				f'📂 {parse_fcnt}|{parse_ftotal} - Loading scan 👷'
			))

			tmp_name = b64decode(tmp_name.split(',')[1]).decode()
			tmp_file = path.join(tempfile.gettempdir(),tmp_name)
			fsize = stat(tmp_file).st_size

			# Let each ingestor validate the file to see if it can parse the scan
			ingestor = None
			for cls in available_ingestors:
				i = cls()
				try:
					if i.validate(tmp_file):
						ingestor = i
						break
				except Exception as e:
					toasts.append(CustomToast(
						[ f'[{i.getName()}] Failed to validate {tmp_file} ({name}): {e}' ],
						headerText=f'Parse error',
						level='error',
						duration=6000,
					))
					# This may have been an error in the parser
					# but another parser could still be valid
					continue

			if ingestor is None:
				toasts.append(CustomToast(
					[ f'Filetype of {name} is not yet supported.' ],
					headerText=f'Missing ingestor',
					level='warning',
					duration=6000,
				))
				continue

			try:
				ingestor.parse()
				# Delete tmp file
				remove(tmp_file)
			except Exception as e:
				toasts.append(CustomToast(
					[ f'Failed to parse {name} ({tmp_file} may still exist): {e}' ],
					headerText=f'Ingestor failed',
					level='warning',
					duration=10000,
				))
				continue

			scan_interface = ingestor.getDatabaseInterface()

			if scan_interface['hosts_scanned'] < 1:
				toasts.append(CustomToast(
					[ f'No hosts were scanned in {name}.' ],
					headerText=f'Scan skipped',
					level='info',
					duration=5000,
				))
				continue

			# Now we parsed the file and can put it into the database

			total_hosts = len(scan_interface['hosts'])

			# Set progress to base percentage + start of nth file + 30% of the assigned percentages per file
			# This represents that the workload per scan is roughly 30% reading, 70% database manipulation
			set_progress((
				parse_bpct+parse_fpct*(parse_fcnt-0.7),
				f'📂 {parse_fcnt}|{parse_ftotal} - Added 0|{total_hosts} 🖥️'
			))

			db_con, db = getDB()

			# Only now we are kind of sure that we want to store the scan, so create a file entry in the db
			db.execute('INSERT INTO input_files(type, filename, ingestDate, filesize) VALUES(?,?,unixepoch(),?)',(
				ingestor.getName(),
				name,
				fsize
			))
			fileID = db.lastrowid

			db.execute('INSERT INTO scans(file, tool, args, version, start, stop, hostsUp, hostsScanned) VALUES (?,?,?,?,?,?,?,?)', (
				fileID,
				scan_interface['tool'],
				scan_interface['args'],
				scan_interface['version'],
				scan_interface['start'],
				scan_interface['stop'],
				scan_interface['hosts_up'],
				scan_interface['hosts_scanned']
			))
			scanID = db.lastrowid

			shosts = scan_interface['hosts']

			added_hosts_cnt = 0
			for shost in shosts:
				added_hosts_cnt += 1 # used for the progress bar at the end

				# All of these addresses are considered unique identifiers for a host:
				# 	IPv4, IPv6, MAC
				# If we detect one of these in the database, we update all new available addresses for the host!
				qr = db.execute('SELECT hid, ipv4, ipv6, mac, reason, os_name, os_accuracy, os_family, os_vendor FROM hosts WHERE (ipv4=?) OR (ipv6=?) OR (mac=?)', (
					IPtoNum(shost['ipv4']),
					shost['ipv6'],
					shost['mac'],
				)).fetchone()
				if qr is not None:
					# Host already exists
					hostID,eipv4,eipv6,emac,ereason,eos_name,eos_accuracy,eos_family,eos_vendor = qr
					# Use the newer IPv4/IPv6/MAC (values may be None if neither the existing nor the new value is set to a value)
					new_ipv4 = (eipv4 if shost['ipv4'] == '' else IPtoNum(shost['ipv4']))
					new_ipv6 = (eipv6 if shost['ipv6'] == '' else shost['ipv6'])
					new_mac = (emac if shost['mac'] == '' else shost['mac'])
					# Generate distinct list of current and new hostnames
					e_hostnames = db.execute('SELECT name FROM hostnames WHERE host=?',(hostID,)).fetchall()
					e_hostnames = [ehn[0] for ehn in e_hostnames]
					new_hostnames = e_hostnames + [nhn for nhn in shost['names'] if nhn not in e_hostnames]
					# Generate the label (replace None with empty strings)
					new_label = getHostLabel(NumToIP(new_ipv4 or -1), new_ipv6 or '', new_mac or '', new_hostnames)

					db.execute('UPDATE hosts SET ipv4=?, ipv6=?, mac=?, reason=?, os_name=?, os_accuracy=?, os_family=?, os_vendor=?, label=? WHERE hid=?', (
						new_ipv4,
						new_ipv6,
						new_mac,
						(ereason if shost['reason'] == '' else shost['reason']),
						# Only update when accuracy is better
						# When the accuracy is better
						(eos_name if int(shost['os_accuracy']) < int(eos_accuracy) or shost['os_name'] == '' else shost['os_name']),
						(eos_accuracy if int(shost['os_accuracy']) < int(eos_accuracy) else shost['os_accuracy']),
						(eos_family if int(shost['os_accuracy']) < int(eos_accuracy) or shost['os_family'] == '' else shost['os_family']),
						(eos_vendor if int(shost['os_accuracy']) < int(eos_accuracy) or shost['os_vendor'] == '' else shost['os_vendor']),
						new_label,
						hostID
					))
				else:
					# Host does not exist yet
					host_label = getHostLabel(shost['ipv4'], shost['ipv6'], shost['mac'], shost['names'])
					db.execute('INSERT INTO hosts(ipv4, ipv6, mac, reason, os_name, os_accuracy, os_family, os_vendor, label) VALUES(?,?,?,?,?,?,?,?,?)', (
						None if ((newIP:=IPtoNum(shost['ipv4'])) < 0) else newIP,
						None if shost['ipv6'] == '' else shost['ipv6'],
						None if shost['mac'] == '' else shost['mac'],
						shost['reason'],
						shost['os_name'],
						shost['os_accuracy'],
						shost['os_family'],
						shost['os_vendor'],
						host_label
					))
					hostID = db.lastrowid

				# Update hostname values
				for new_name in shost['names']:
					# doubles (produced by nmap for user and PTR type for example) will be ignored by the DB constraint UNIQUE
					db.execute('INSERT OR IGNORE INTO hostnames(host, name) VALUES (?, ?)', (
						hostID,
						new_name
					))

				try:
					# If this fails due to a UNIQUE constraint, then
					# the scan contains two results for one host.
					db.execute('INSERT INTO scans_hosts(scan, host) VALUES(?,?)', (
						scanID,
						hostID
					))
				except:
					# This should not happen but we updated the host already and simply
					# skip the ports - let's continue after that.
					toasts.append(CustomToast(
						[ f'Found two results for one host in the same scan: {name}.' ],
						headerText=f'Scan possibly malformed',
						level='warning',
						duration=6000,
					))
					continue

				hports = shost['ports']
				for hport in hports:
					qr = db.execute('SELECT pid, svc_name, svc_info, svc_ssl FROM ports WHERE host=? AND port=? AND protocol=?', (
						hostID,
						hport['port'],
						hport['protocol']
					)).fetchone()
					if qr is not None:
						# Port already exists
						portID,esvc_name,esvc_info,esvc_ssl = qr
						db.execute('UPDATE ports SET svc_name=?, svc_info=?, svc_ssl=? WHERE pid=?',(
							# Use the new value if one exists
							(esvc_name if hport['svc_name'] == '' else hport['svc_name']),
							(esvc_info if hport['svc_info'] == '' else hport['svc_info']),
							(esvc_ssl if hport['svc_ssl'] == '' else hport['svc_ssl']),
							portID
						))
					else:
						# Port does not exist yet
						db.execute('INSERT INTO ports(host, port, protocol, svc_name, svc_info, svc_ssl) VALUES(?,?,?,?,?,?)', (
							hostID,
							hport['port'],
							hport['protocol'],
							hport['svc_name'],
							hport['svc_info'],
							hport['svc_ssl'],
						))
						portID = db.lastrowid

					db.execute('INSERT INTO scans_ports(scan, port) VALUES(?,?)', (
						scanID,
						portID
					))

					pscripts = hport['scripts']
					for pscript in pscripts:
						# If the same script was already run by a previous scan, then overwrite the results
						# Reason: A second scan in the network may reveal new results and we only want the up to date information
						qr = db.execute('SELECT psid FROM pscripts WHERE port=? AND host=? AND name=?', (portID, hostID, pscript['id'])).fetchone()
						if qr is not None:
							pscriptID = qr[0]
							db.execute('UPDATE pscripts SET scan=?, output=? WHERE psid=?', (
								scanID, pscript['output'], pscriptID # Keep in mind that this pscriptID is not pscript['id']! One is the database primary key, the other the script name.
							))
						else:
							db.execute('INSERT INTO pscripts(scan, host, port, name, output) VALUES(?,?,?,?,?)', (
								scanID, hostID, portID, pscript['id'], pscript['output']
							))

				# Fill up the remaining 70% of every scan with the current amount of processed hosts
				# 70%
				set_progress((
					parse_bpct+parse_fpct*(parse_fcnt-0.7 + 0.7*added_hosts_cnt/total_hosts),
					f'📂 {parse_fcnt}|{parse_ftotal} - Added {added_hosts_cnt}|{total_hosts} 🖥️'
				))

			toasts.append(CustomToast(
				[ f'Found {scan_interface["hosts_up"]} hosts in {name}.' ],
				headerText=f'Scan imported',
				level='success',
				duration=3000,
			))

			# Be sparse with transactions as they take longer than statements (because changes are actually written to disk)
			# https://www.sqlite.org/faq.html#q19
			db_con.commit()
			db_con.close()
		
		# All scans processed
		set_progress((100, f'All done 🎉'))
		time.sleep(2) # Keep this for better UX

	return True, toasts


@callback(	Output('input-files-table', 'rowData'),
		Input('upload-data-change', 'data'),
	)
def _cb_updateFileUploadTable(_):
	db_con, db = getDB()
	# Nested SQL queries won't work with the same connection when iterating over a cursor
	# Alternatively we'd have to use fetchall()
	db_con_secondary, db_secondary = getDB()

	if_table_rows = []
	qr = db.execute('SELECT fid, type, filename, datetime(ingestDate,"unixepoch","localtime"), filesize FROM input_files')
	for file in qr:
		fid = file[0]
		hosts_scanned, hosts_up = db_secondary.execute('SELECT hostsScanned, hostsUp FROM scans WHERE file=?', (fid,)).fetchone()
		if_table_rows.append(dict(
			type=file[1],
			filename=file[2],
			targetsScanned=hosts_scanned,
			targetsUp=hosts_up,
			ingestDate=file[3],
			filesize=file[4]
		))
	db_con.close()
	db_con_secondary.close()
	return if_table_rows


@callback(	Output('modal-confirm-delete-scans', 'is_open'),
		Input('btn-delete-scans', 'n_clicks'),
		State('input-files-table', 'rowData'),
		prevent_initial_call = True
	)
def _cb_deleteAllScans(n, eles):
	return (len(eles)>0)

@callback( 	Output('upload-data-change', 'data', allow_duplicate=True),
		Output('modal-confirm-delete-scans', 'is_open', allow_duplicate=True),
		Input('btn-confirm-delete-scans', 'n_clicks'),
		prevent_initial_call = True
)
def _cb_confirmDeleteAllScans(n_clicks):
	if n_clicks:
		con, cur = getDB()
		cur.execute('DELETE FROM scans_ports')
		cur.execute('DELETE FROM scans_hosts')
		cur.execute('DELETE FROM ports')
		cur.execute('DELETE FROM pscripts')
		cur.execute('DELETE FROM hosts')
		cur.execute('DELETE FROM hostnames')
		cur.execute('DELETE FROM scans')
		cur.execute('DELETE FROM input_files')
		con.commit()
		con.close()
	return True, False



from dash import html, get_asset_url
import dash_bootstrap_components as dbc

# There appears to be a glitch that may rerender a previous toast when appending new toasts
# using the previous children of the toast div.
# It's not critical for now, so we leave it as is.
class CustomToast(dbc.Toast):
	def __init__(self,eles, headerText, level='info', duration=4000):

		if level not in ['info', 'warning', 'error', 'success']:
			level = 'info'

		header = [
			html.Img(src=get_asset_url(f'icons/toast/{level}.svg')),
			html.Span(headerText, style={'marginLeft': '0.5rem'})
		]

		super().__init__(
			eles,
			header=header,
			is_open=True,
			dismissable=True,
			duration=duration,
		)



#!/usr/bin/env python3

from scandeavour.ingestor_base import BaseIngestor
import sys
import os
import xml.etree.ElementTree as ET

class fc:
	red = '\u001b[31m'
	green = '\u001b[32m'
	orange = '\u001b[33m'
	blue = '\u001b[34m'
	end = '\u001b[0m'

class NmapIngestor(BaseIngestor):
	def __init__(self):
		super().__init__(
			name = 'NmapXML',
			accepted_files = 'nmap.xml'
		)

	def validate(self, file_path):
		# Other tools such as masscan may use a similar format, so be sure to check this correctly
		with open(file_path, 'rb') as f:
			# line 1
			line = f.readline()
			if not line.startswith(b'<?xml version'):
				return False
			# line 2
			line = f.readline()
			if not line.startswith(b'<!DOCTYPE nmaprun'): 
				return False
			
			self.file_path = file_path
			return True

	def parse(self):
		self.root = None
		try:
			self.root = ET.parse(self.file_path).getroot()
		except Exception as e:
			print(f'{fc.red}[!]{fc.end} Parsing failed with: {e}')
			sys.exit(1)

		nmaprunner = self.root.attrib
		self.cmdline = nmaprunner.get('args','') # optional
		self.scanstart = int(nmaprunner.get('start',0)) # unix timestamp - optional
		self.nmapversion = nmaprunner['version'] # mandatory

		runstats_finished = self.root.find('runstats/finished').attrib # mandatory
		self.scanstop = int(runstats_finished['time']) # unix timestamp - mandatory

		runstats_hosts = self.root.find('runstats/hosts').attrib # mandatory
		self.hosts_up = int(runstats_hosts['up']) # mandatory
		self.hosts_down = int(runstats_hosts['down']) # mandatory
		self.hosts_total = int(runstats_hosts['total']) # mandatory

		self.hosts = []
		xml_hosts = self.root.findall('host') # optional
		for xml_host in xml_hosts:
			host_status = xml_host.find('status').attrib # mandatory
			if host_status['state'] != 'up': # mandatory - up|down|unknown|skipped
				continue
			new_host = {}
			new_host['up_reason'] = host_status['reason'] # mandatory

			host_addrs = xml_host.findall('address') # mandatory - at least one

			addrs = []
			for host_addr in host_addrs:
				attribs = host_addr.attrib
				addrs.append({
					'type': attribs['addrtype'], # ipv4|ipv6|mac - mandatory
					'addr': attribs['addr'] # mandatory
				})
			new_host['addrs'] = addrs

			host_names = xml_host.findall('hostnames/hostname') # optional
			names = []
			for name in host_names:
				names.append(name.attrib.get('name','')) # optional

			new_host['hostnames'] = names

			xml_os = xml_host.find('os/osmatch') # optional
			if xml_os is not None:
				match = xml_os.attrib
				new_host['accuracy'] = int(match['accuracy']) # mandatory
				new_host['osname'] = match['name'] # mandatory
			else:
				new_host['accuracy'] = 0
				new_host['osname'] = ''

			osclass = xml_host.find('os/osmatch/osclass') # optional
			if xml_os is not None and osclass is not None:
				new_host['osfamily'] = osclass.attrib['osfamily'] # mandatory
				new_host['vendor'] = osclass.attrib['vendor'] # mandatory
			else:
				new_host['osfamily'] = ''
				new_host['vendor'] = ''

			# parse ports
			xml_ports = xml_host.findall('ports/port') # optional
			ports = []
			for xml_port in xml_ports:
				port = {}
				port_state = xml_port.find('state').attrib # mandatory
				# State can be one of:
				# open, filtered, unfiltered, closed, open|filtered, closed|filtered, unknown
				port['state'] = port_state['state'] # mandatory
				if port['state'] in ['filtered', 'closed', 'closed|filtered']:
					# ignore closed ports
					continue

				port['protocol'] = xml_port.attrib['protocol'] # mandatory - one of ip, udp, tcp, sctp
				port['number'] = int(xml_port.attrib['portid']) # mandatory

				xml_svc = xml_port.find('service') # optional
				xml_svc = xml_svc.attrib if (xml_svc is not None) else {}
				port['svc_name'] = xml_svc.get('name', '') # name
				port['svc_conf'] = int(xml_svc.get('conf', 0)) # confidence (value between 0 and 10)
				port['svc_product'] = xml_svc.get('product', '') # product
				port['svc_version'] = xml_svc.get('version', '') # version
				port['svc_extrainfo'] = xml_svc.get('extrainfo', '') # additional information
				port['svc_tunnel'] = xml_svc.get('tunnel', '') # optional - if set, then set to "ssl"
				port['svc_proto'] = xml_svc.get('proto', '') # optional - if set, then set to "rpc"

				xml_port_scripts = xml_port.findall('script') # optional
				port_scripts = []
				for xml_port_script in xml_port_scripts:
					port_scripts.append({
						'id': xml_port_script.get('id'), # mandatory
						'output': xml_port_script.get('output').strip('\n'), # mandatory
					})
				port['scripts'] = port_scripts

				ports.append(port)

			new_host['ports'] = ports

			self.hosts.append(new_host)

	# Assume that every host can only have one IPv4, IPv6 and MAC
	def getAddrByType(self, addrs, type):
		for addr in addrs:
			if addr['type'] == type:
				return addr['addr']
		return ''

	def getDatabaseInterface(self):
		scan = {
			'tool': 'Nmap',
			'args': self.cmdline,
			'start': self.scanstart,
			'stop': self.scanstop,
			'version': self.nmapversion,
			'hosts_up': self.hosts_up,
			'hosts_scanned': self.hosts_total,
			'hosts': [{
				'ipv4': self.getAddrByType(host['addrs'], 'ipv4'),
				'ipv6': self.getAddrByType(host['addrs'], 'ipv6'),
				'mac': self.getAddrByType(host['addrs'], 'mac'),
				'names': host['hostnames'], # array of names
				'reason': host['up_reason'],
				'os_name': host['osname'],
				'os_accuracy': host['accuracy'],
				'os_family': host['osfamily'],
				'os_vendor': host['vendor'],
				'ports': [{
					'port': port['number'],
					'protocol': port['protocol'],
					'svc_name': port['svc_name'],
					'svc_info': f'{port["svc_product"]} {port["svc_version"]} {port["svc_extrainfo"]}'.strip(),
					'svc_ssl': port['svc_tunnel'],
					'scripts': port['scripts']
				} for port in host["ports"]],
			} for host in self.hosts]
		}
		return scan


	def __str__(self):
		str = ''
		str += f'CMD: {self.cmdline}\n'
		str += f'Scan started: {self.scanstart}\n'
		str += f'Scan stopped: {self.scanstop}\n'
		str += f'Nmap version: {self.nmapversion}\n'
		str += f'Scanned hosts: {self.hosts_up} up, {self.hosts_down} down, {self.hosts_total} total\n'
		for host in self.hosts:
			str += f'====\n'
			str += f'\tAddrs: '+ ', '.join(((a['addr']+" ("+a['type']+")") for a in host['addrs'])) + '\n'
			str += f'\tNames: '+ ', '.join(host['hostnames']) + '\n'
			str += f'\tUp-Reason: {host["up_reason"]}\n'
			str += f'\tOS name: {host["osname"]}\n'
			str += f'\tOS accuracy: {host["accuracy"]}\n'
			str += f'\tOS family: {host["osfamily"]}\n'
			str += f'\tOS vendor: {host["vendor"]}\n'
			str += f'\tPorts:\n'
			for port in host['ports']:
				str += f'\t\t{port["number"]:5}/{port["protocol"]} '
				str += f'{port["svc_name"]} ({port["svc_conf"]}|10) '
				str += f'{port["svc_product"]} '
				str += f'{port["svc_version"]} '
				str += f'{port["svc_extrainfo"]} '
				str += f'{port["svc_tunnel"]} '
				str += f'{port["svc_proto"]}\n'
				for script in port['scripts']:
					str += f'{script["id"]}: {script["output"]}\n'
		return str

# Standalone version
# Requires a copy of ingestor_base.py in the same directory
if __name__ == '__main__':
	if len(sys.argv) != 2:
		print(f'Usage: {sys.argv[0]} nmap_scan_output.xml')
		sys.exit(1)

	inp = sys.argv[1]
	if not os.path.isfile(inp):
		print(f'{fc.red}[!]{fc.end} {inp} does not exist')
		sys.exit(1)

	ing = NmapIngestor()
	if not ing.validate(inp):
		print(f'{fc.red}[!]{fc.end} Not a valid Nmap file')
		sys.exit(1)
	ing.parse()
	print(ing)



#!/usr/bin/env python3

# Based on the sourcecode from masscan/src/out-xml.c.
# The current masscan XML output is flawed (for example "hosts up" shows the port number not host number) and may change in the future.
# So when things start to break, look at the changes made to out-xml.c in masscan.

from scandeavour.ingestor_base import BaseIngestor
import sys
import os
import xml.etree.ElementTree as ET

class fc:
	red = '\u001b[31m'
	green = '\u001b[32m'
	orange = '\u001b[33m'
	blue = '\u001b[34m'
	end = '\u001b[0m'

class MasscanXMLIngestor(BaseIngestor):
	def __init__(self):
		super().__init__(
			name = 'MasscanXML',
			accepted_files = 'masscan.xml'
		)

	def validate(self, file_path):
		# masscan XML should start with XML that contains a comment like "masscan"
		# see https://github.com/robertdavidgraham/masscan/blob/master/src/out-xml.c
		with open(file_path, 'rb') as f:
			# line 1
			line = f.readline()
			if not line.startswith(b'<?xml version'):
				return False
			# line 2
			line = f.readline()
			if b'<!-- masscan' not in line: 
				return False
			
			self.file_path = file_path
			return True

	def parse(self):
		self.root = None
		try:
			self.root = ET.parse(self.file_path).getroot()
		except Exception as e:
			print(f'{fc.red}[!]{fc.end} Parsing failed with: {e}')
			sys.exit(1)

		msrunner = self.root.attrib
		self.scanstart = int(msrunner.get('start',0)) # unix timestamp
		# can't use the version because it is hardcoded in out-xml.c and not updated since the beta release
		#self.msversion = msrunner['version']

		runstats_finished = self.root.find('runstats/finished').attrib # mandatory
		self.scanstop = int(runstats_finished['time']) # unix timestamp - mandatory

		# can't use the runstats as they currently count open ports not the hosts
		#runstats_hosts = self.root.find('runstats/hosts').attrib # mandatory
		#self.hosts_up = int(runstats_hosts['up']) # mandatory
		#self.hosts_total = int(runstats_hosts['total']) # mandatory

		self.hosts = {}
		xml_hosts = self.root.findall('host') # optional
		for xml_host in xml_hosts:

			# can't rely on the addrtype for the current masscan release, see https://github.com/robertdavidgraham/masscan/issues/807
			host_addr = xml_host.find('address').attrib.get('addr') # masscan only shows one addr

			if host_addr not in self.hosts.keys():
				self.hosts[host_addr] = {
					'ipv6': '',
					'ipv4': '',
					'ports': []
				}

			if ':' in host_addr:
				self.hosts[host_addr]['ipv6'] = host_addr
			else:
				self.hosts[host_addr]['ipv4'] = host_addr

			# currently, masscan XML shows one port per host, so a host with multiple ports will actually show up separately for every open port
			xml_port = xml_host.find('ports/port') # mandatory
			port_state = xml_port.find('state').attrib.get('state') # mandatory

			port_protocol = xml_port.attrib['protocol'] # mandatory
			port_number = int(xml_port.attrib['portid']) # mandatory

			if 'open' in port_state:
				# I assume that, like nmap, masscan will show a port only once for every host
				self.hosts[host_addr]['ports'].append({
					'number': port_number,
					'protocol': port_protocol
				})
			else:
				print(f'{fc.red}[!]{fc.end} Masscan identified {host_addr} - {port_number}/{port_protocol}. Host added but port skipped due to state: {port_state}')
				continue

	def getDatabaseInterface(self):
		scan = {
			'tool': 'Masscan',
			'args': '', # not implemented by masscan
			'start': self.scanstart,
			'stop': self.scanstop,
			'version': '', # not implemented by masscan
			'hosts_up': len(self.hosts),
			'hosts_scanned': len(self.hosts), # not implemented by masscan
			'hosts': [{
				'ipv4': host['ipv4'],
				'ipv6': host['ipv6'],
				'mac': '', # not implemented by masscan
				'names': [], # not implemented by masscan
				'reason': '', # not implemented by masscan
				'os_name': '', # not implemented by masscan
				'os_accuracy': 0, # not implemeneted by masscan
				'os_family': '', # not implemented by masscan
				'os_vendor': '', # not implemented by masscan
				'ports': [{
					'port': port['number'],
					'protocol': port['protocol'],
					'svc_name': '', # not implemented
					'svc_info': '', # not implemented
					'svc_ssl': '', # not implemented
					'scripts': [] # not implemented
				} for port in host["ports"]],
			} for host in self.hosts.values()]
		}
		return scan


	def __str__(self):
		str = ''
#		str += f'CMD: {self.cmdline}\n'
		str += f'Scan started: {self.scanstart}\n'
		str += f'Scan stopped: {self.scanstop}\n'
#		str += f'Masscan version: {self.msversion}\n'
		str += f'Scanned hosts: {len(self.hosts)}\n'
		for host in self.hosts:
			str += f'====\n'
			str += f'\tAddr: v4: '+host['ipv4']+', v6: '+host['ipv6']+'\n'
#			str += f'\tNames: '+ ', '.join(host['hostnames']) + '\n'
#			str += f'\tUp-Reason: {host["up_reason"]}\n'
#			str += f'\tOS name: {host["osname"]}\n'
#			str += f'\tOS accuracy: {host["accuracy"]}\n'
#			str += f'\tOS family: {host["osfamily"]}\n'
#			str += f'\tOS vendor: {host["vendor"]}\n'
			str += f'\tPorts:\n'
			for port in host['ports']:
				str += f'\t\t{port["number"]:5}/{port["protocol"]}\n'
#				str += f'{port["svc_name"]} ({port["svc_conf"]}|10) '
#				str += f'{port["svc_product"]} '
#				str += f'{port["svc_version"]} '
#				str += f'{port["svc_extrainfo"]} '
#				str += f'{port["svc_tunnel"]} '
#				str += f'{port["svc_proto"]}\n'
#				for script in port['scripts']:
#					str += f'{script["id"]}: {script["output"]}\n'
		return str

# Standalone version
# Requires a copy of ingestor_base.py in the same directory
if __name__ == '__main__':
	if len(sys.argv) != 2:
		print(f'Usage: {sys.argv[0]} masscan_scan_output.xml')
		sys.exit(1)

	inp = sys.argv[1]
	if not os.path.isfile(inp):
		print(f'{fc.red}[!]{fc.end} {inp} does not exist')
		sys.exit(1)

	ing = MasscanIngestor()
	if not ing.validate(inp):
		print(f'{fc.red}[!]{fc.end} Not a valid Masscan file')
		sys.exit(1)
	ing.parse()
	print(ing)



from dash import html, callback, Output, Input, State, dcc, register_page, get_asset_url, ctx
from dash.exceptions import PreventUpdate
import plotly.graph_objects as go
import dash_ag_grid as dag
import dash_bootstrap_components as dbc
from scandeavour.utils import getDB
from scandeavour.components.customToast import CustomToast

register_page(__name__, path='/stats')

def layout(**kwargs):
	db_con, db = getDB()

	default_value = '-/-'

	# Get total ports count
	total_ports = 0
	qr = db.execute('SELECT COUNT(port||protocol) AS cnt FROM ports').fetchone()
	if qr is not None:
		total_ports = qr[0]

	# Get most common ports
	qr = db.execute('SELECT port || "/" || protocol, COUNT(*) AS cnt FROM ports GROUP BY port||protocol ORDER BY cnt DESC LIMIT 4').fetchall()

	most_common_ports = [
		{ 'port': default_value, 'count': 0 },
		{ 'port': default_value, 'count': 0 },
		{ 'port': default_value, 'count': 0 },
		{ 'port': default_value, 'count': 0 },
	]
	i = 0
	for r in qr:
		most_common_ports[i]['port'] = r[0]
		most_common_ports[i]['count'] = r[1]
		i += 1

	# Get least common ports
	qr = db.execute('SELECT port || "/" || protocol, COUNT(*) AS cnt FROM ports GROUP BY port||protocol ORDER BY cnt ASC LIMIT 4').fetchall()

	least_common_ports = [
		{ 'port': default_value, 'count': 0 },
		{ 'port': default_value, 'count': 0 },
		{ 'port': default_value, 'count': 0 },
		{ 'port': default_value, 'count': 0 },
	]
	i = 0
	for r in qr:
		least_common_ports[i]['port'] = r[0]
		least_common_ports[i]['count'] = r[1]
		i += 1

	db_con.close()

	return html.Div([
		dcc.Interval(
			id='interval-once-stats',
			n_intervals=0,
			max_intervals=0,
			interval=1
		),
		dcc.Clipboard(id='stats-clipboard', style={'display': 'none'}, n_clicks=0),
		html.Div([
			html.Div([
				html.Div([
					html.Img(src=get_asset_url('icons/graphpage/box.svg')),
					html.H4('Most common ports')
				],
				className='medium-heading'),
				html.Div(
				[
					html.Div([
						html.Div([
							html.Div([], className='gauge-background'),
							html.Div([], className='gauge-filler', style={'transform': f'rotate({int((int(most_common_ports[0]["count"])*180)/total_ports) if total_ports > 0 else 0}deg)'}),
							html.Div([
								html.H5(most_common_ports[0]['port'])
							], className='gauge-data')
						],
						className='gauge-container'),
						html.Span(f'{most_common_ports[0]["count"]} | {total_ports}', className='gauge-label')
					]),
					dbc.ListGroup(
						[
							dbc.ListGroupItem(html.Div([html.Span(most_common_ports[i]['port']), html.Span(most_common_ports[i]['count'])], className='lg-key-value'))
							for i in range(1,4)
						],
						flush=True,
						className='gauge-data-table'
					),
				],
				className='graph-stats-wrapper')
			],
			className='stats-box'),
			html.Div([
				html.Div([
					html.Img(src=get_asset_url('icons/graphpage/box.svg')),
					html.H4('Least common ports')
				],
				className='medium-heading'),
				html.Div(
				[
					html.Div([
						html.Div([
							html.Div([], className='gauge-background'),
							html.Div([], className='gauge-filler', style={'transform': f'rotate({int((int(least_common_ports[0]["count"])*180)/total_ports) if total_ports > 0 else 0}deg)'}),
							html.Div([
								html.H5(least_common_ports[0]['port'])
							], className='gauge-data')
						],
						className='gauge-container'),
						html.Span(f'{least_common_ports[0]["count"]} | {total_ports}', className='gauge-label')
					]),
					dbc.ListGroup(
						[
							dbc.ListGroupItem(html.Div([html.Span(least_common_ports[i]['port']), html.Span(least_common_ports[i]['count'])], className='lg-key-value'))
							for i in range(1,4)
						],
						flush=True,
						className='gauge-data-table'
					),
				],
				className='graph-stats-wrapper')
			],
			className='stats-box'),
			html.Div([
				html.Div([
					html.Img(src=get_asset_url('icons/headings/tool.svg')),
					html.H4('Identified services'),
				],
				className='medium-heading'),
				dcc.Loading([
						html.Div([
							dcc.Graph(id='stats-service-figure', figure=go.Figure(), className='stats-graph-container'),
							dag.AgGrid(
								id='stats-service-table',
								rowData=[],
								columnDefs=[
									{'headerName': 'Service', 'field': 'svc', 'maxWidth': '100'},
									{'headerName': 'Details', 'field': 'svc_details', 'flex': 1},
									{'headerName': 'Count', 'field': 'count', 'sort': 'desc', 'maxWidth': '100'},
								],
								defaultColDef={
									'wrapText': True,
									'autoHeight':True,
									'cellStyle': {'wordBreak':'normal'},
								},
								className='ag-theme-alpine-dark stats-ag-grid',
								style={'height': '16rem'}
							)
						],
						className='graph-stats-wrapper')
					],
					overlay_style={'visibility': 'visible', 'filter': 'blur(2px)'},
					type='circle',
					delay_hide=200,
					color='#00ffe0'
				)
			],
			className='stats-box grid-span-two'),
			html.Div([
				html.Div([
					html.Img(src=get_asset_url('icons/headings/cpu.svg')),
					html.H4('Operating systems')
				],
				className='medium-heading'),
				dcc.Loading([
						html.Div([
							dcc.Graph(id='stats-os-figure', figure=go.Figure(), className='stats-graph-container'),
							dag.AgGrid(
								id='stats-os-table',
								rowData=[],
								columnDefs=[
									{'headerName': 'Family', 'field': 'os_fam', 'maxWidth': '100'},
									{'headerName': 'Vendor', 'field': 'os_ven', 'flex': 1},
									{'headerName': 'Name', 'field': 'os_nam', 'flex': 1},
									{'headerName': 'Count', 'field': 'count', 'sort': 'desc', 'maxWidth': '100'},
								],
								defaultColDef={
									'wrapText': True,
									'autoHeight':True,
									'cellStyle': {'wordBreak':'normal'},
								},
								className='ag-theme-alpine-dark stats-ag-grid',
								style={'height': '16rem'}
							)
						],
						className='graph-stats-wrapper')
					],
					overlay_style={'visibility': 'visible', 'filter': 'blur(2px)'},
					type='circle',
					delay_hide=200,
					color='#00ffe0'
				)
			],
			className='stats-box grid-span-two'),
		],
		className='stats-grid')
	],
	className='graph-container')


@callback(	Output('stats-clipboard', 'content'),
		Output('div-toaster', 'children', allow_duplicate=True),
		Output('stats-clipboard', 'n_clicks'),
		Input('stats-service-table', 'cellDoubleClicked'),
		Input('stats-os-table', 'cellDoubleClicked'),
		State('div-toaster', 'children'),
		State('stats-clipboard', 'n_clicks'),
		prevent_initial_call=True
)
def _cb_copy_table_cell_to_clipboard(svc_cell, os_cell, toasts, n_clipboard):
	cell = None
	if ctx.triggered_id == 'stats-service-table':
		cell = svc_cell
	elif ctx.triggered_id == 'stats-os-table':
		cell = os_cell

	if cell is None or 'value' not in cell:
		raise PreventUpdate

	clipboard_value = cell['value']

	toasts.append(CustomToast(
		[
			'Copied cell content to clipboard.'
		],
		duration = 1000,
		headerText = 'Value copied',
		level = 'success'
	))

	n_clipboard += 1
	return clipboard_value, toasts, n_clipboard


@callback(	Output('stats-service-table', 'rowData'),
		Output('stats-service-figure', 'figure'),
		Output('stats-os-table', 'rowData'),
		Output('stats-os-figure', 'figure'),
		Input('stats-service-table', 'rowData'), # dummy Input to trigger callback
	)
def _cb_init_stats_page(_):
	db_con, db = getDB()

	# Get all services
	qr = db.execute('SELECT svc_name, svc_info, COUNT(*) AS cnt FROM ports GROUP BY svc_name||svc_info ORDER BY cnt DESC').fetchall()

	svc_labels = []
	svc_details = []
	svc_values = []
	svc_combo = []

	for r in qr:
		svc_labels.append(r[0])
		svc_details.append(r[1])
		svc_values.append(int(r[2]))
		svc_combo.append(f'{r[0] if r[0]!="" else "-/-"}{" - " if r[1]!="" else ""}{r[1]}')

	svc_table_data = [{'svc': svc_labels[i], 'svc_details': svc_details[i], 'count': svc_values[i]} for i in range(len(svc_labels))]

	defaultLayout = go.Layout(
		paper_bgcolor='rgba(0,0,0,0)',
		plot_bgcolor='rgba(0,0,0,0)',
		margin = go.layout.Margin(
			l = 0,
			r = 0,
			b = 0,
			t = 0
		),
		showlegend = False,
		uniformtext_minsize=12,
		uniformtext_mode='hide',
		colorway = [
			'#4E878C',
			'#508B8D',
			'#518E8D',
			'#53918E',
			'#54948E',
			'#56978F',
			'#579A8F',
			'#599D8F',
			'#599D8F',
			'#5AA08F',
			'#60AC90',
			'#63B291',
			'#65B891',
			'#6BBE95',
			'#71C498',
			'#77CA9B',
			'#7CCF9E',
			'#82D5A2',
			'#88DAA5',
		]
	)

	# Pie chart for services
	svc_donut = go.Figure(
		data = [go.Pie(
			labels=svc_combo,
			values=svc_values,
			hole=.5,
			textposition='inside',
			direction='clockwise',
			sort=True,
		)],
		layout = defaultLayout
	)


	# Get all Operating systems
	qr = db.execute('SELECT os_family, os_vendor, os_name, COUNT(*) AS cnt FROM hosts GROUP BY os_family||os_vendor||os_name ORDER BY cnt DESC').fetchall()

	os_family = []
	os_vendor = []
	os_name = []
	os_combo = []
	os_values = []

	for r in qr:
		os_family.append(r[0])
		os_vendor.append(r[1])
		os_name.append(r[2])
		os_combo.append(f'{r[0] if r[0]!="" else "-"} / {r[1] if r[1]!="" else "-"} / {r[2] if r[2]!="" else "-"}')
		os_values.append(int(r[3]))

	os_table_data = [{'os_fam': os_family[i], 'os_ven': os_vendor[i], 'os_nam': os_name[i], 'count': os_values[i]} for i in range(len(os_name))]
	# Pie chart for OS
	os_donut = go.Figure(
		data = [go.Pie(
			labels=os_combo,
			values=os_values,
			hole=.5,
			textposition='inside',
			direction='clockwise',
			sort=True,
		)],
		layout = defaultLayout
	)

	db_con.close()

	return svc_table_data, (svc_donut), os_table_data, (os_donut)



from dash import html, dcc, callback, Input, Output, State, register_page, get_asset_url, ctx, no_update, ALL, Patch
import dash_bootstrap_components as dbc
import dash_cytoscape as cyto
from scandeavour.components.customToast import CustomToast
from scandeavour.utils import getDB, NodeHistory, NumToIP, IPtoNum, CIDRtoFirstLast, validateIP, getOS, DataFilterMap

register_page(__name__, path='/graph')

def layout(**kwargs):
	graph_stylesheet = [
		# Activity
		{
			'selector': ':selected',
			'style': {
				'border-width': '3px',
				'border-color': '#fff',
			}
		},
		# Groups
		{
			'selector': 'label',
			'style': {
				'content': 'data(label)',
				'color': '#00FFE0'
			}
		},
		{
			'selector': 'node',
			'style': {
				'background-opacity': '0',
			}
		},
		{
			'selector': 'edge',
			'style': {
				'curve-style': 'straight-triangle'
			}
		},
		# Classes
		{
			'selector': '.scan',
			'style': {
				'shape': 'round-rectangle',
				'background-opacity': '0.2',
				'background-color': 'black',
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=command
				'background-image': get_asset_url('icons/graphpage/terminal.svg')
			}
		},
		{
			# Options taken from: https://github.com/nmap/nmap/blob/fef9f592b067e53c58c5c593e8916cd83e4bb873/zenmap/zenmapGUI/Icons.py#L80
			'selector': '.linuxhost',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				# & boxy-svg.com
				'background-image': get_asset_url('icons/hosts/host-linux.svg')
			}
		},
		{
			'selector': '.windowshost',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				# & boxy-svg.com
				'background-image': get_asset_url('icons/hosts/host-windows.svg')
			}
		},
		{
			'selector': '.applehost',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				# & boxy-svg.com
				'background-image': get_asset_url('icons/hosts/host-apple.svg')
			}
		},
		{
			'selector': '.redhathost',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				# & boxy-svg.com
				'background-image': get_asset_url('icons/hosts/host-redhat.svg')
			}
		},
		{
			'selector': '.ubuntuhost',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				# & boxy-svg.com
				'background-image': get_asset_url('icons/hosts/host-ubuntu.svg')
			}
		},
		{
			'selector': '.freebsdhost',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				# & boxy-svg.com
				'background-image': get_asset_url('icons/hosts/host-freebsd.svg')
			}
		},
		{
			'selector': '.openbsdhost',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				# & boxy-svg.com
				'background-image': get_asset_url('icons/hosts/host-openbsd.svg')
			}
		},
		{
			'selector': '.solarishost',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				# & boxy-svg.com
				'background-image': get_asset_url('icons/hosts/host-solaris.svg')
			}
		},
		{
			'selector': '.host',
			'style': {
				'width': 70,
				'height': 70,
				# https://feathericons.com/?query=monitor
				'background-image': get_asset_url('icons/graphpage/host.svg'),
			}
		},
		{
			'selector': '.port',
			'style': {
				# https://feathericons.com/?query=box
				'width': 30,
				'height': 30,
				'background-image': get_asset_url('icons/graphpage/box.svg'),
			}
		}
	]

	return html.Div([
		dcc.Store('graph-settings', data={
	                        'scans-visible': True,
	                        'ports-visible': True,
	                        'hostsnp-visible': True,
	                        'datafilter-linked': True,
	                        'filters': {
	                                'subnets': [],
					'datatable': []
	                        },
	                        'max_elements_render': 512,
			}, storage_type='session'),
		dcc.Store('render-trigger', data={
				'trigger': False
			}, storage_type='memory'),
		dcc.Store('click-intent', data={
				'id': ''
			}, storage_type='memory'),
		dcc.Store('node-nav-state', data={
				'history': [],
				'index': -1,
				'jumptable': {},
			}, storage_type='session'),
		cyto.Cytoscape(
			id='scan-graph',
			# Default to very responsive draggable layout for small datasets
			layout = {
				'name': 'cola',
				'animate': True,
				'refresh': 1,
				'avoidOverlap': True,
				'infinite': True,
				'fit': False,
				'nodeDimensionsIncludeLabels': True
			},
			stylesheet=graph_stylesheet,
			style={
				'width':'100%',
				'height': '100vh',
			},
			elements=[],
			responsive=True,
			zoomingEnabled=True,
			zoom=2,
			wheelSensitivity=0.4,
		),
		html.Div([
			html.Div(
				[
					html.Img(
						src=get_asset_url('icons/graphpage/maximize.svg'),
					),
					# Currently, Popovers generate a React warning in the console when we hover over the target the first time after ctrl+f5
					dbc.Popover(
						'Fit view',
						target='btn-fit-graph-view',
						body=True,
						trigger="hover",
					),
				],
				className='graph-overlay-item',
				id='btn-fit-graph-view',
				n_clicks=0,
			),
			html.Div(
				[
					html.Img(
						src=get_asset_url('icons/graphpage/crosshair.svg'),
					),
					dbc.Popover(
						'Center on node',
						target='btn-center-selected-node',
						body=True,
						trigger="hover",
					),
				],
				className='graph-overlay-item',
				id='btn-center-selected-node',
				n_clicks=0,
			),
			html.Div(
				[
					html.Img(
						src=get_asset_url('icons/graphpage/terminal-on.svg'),
						id='toggle-scans-img'
					),
					dbc.Popover(
						'Hide scans',
						target='btn-toggle-scans',
						body=True,
						trigger="hover",
						id='btn-toggle-scans-label',
					),
				],
				className='graph-overlay-item',
				id="btn-toggle-scans",
				n_clicks=0,
			),
			html.Div(
				[
					html.Img(
						src=get_asset_url('icons/graphpage/box-on.svg'),
						id='toggle-ports-img'
					),
					dbc.Popover(
						'Hide ports',
						target='btn-toggle-ports',
						body=True,
						trigger="hover",
						id='btn-toggle-ports-label',
					),
				],
				className='graph-overlay-item',
				id="btn-toggle-ports",
				n_clicks=0,
			),
			html.Div(
				[
					html.Img(
						src=get_asset_url('icons/graphpage/host-on.svg'),
						id='toggle-hosts-img'
					),
					dbc.Popover(
						'Hide hosts with no ports',
						target='btn-toggle-hosts',
						body=True,
						trigger="hover",
						id='btn-toggle-hosts-label',
					),
				],
				className='graph-overlay-item',
				id="btn-toggle-hosts",
				n_clicks=0,
			),
			html.Div(
				[
					html.Img(
						src=get_asset_url('icons/graphpage/datafilter-on.svg'),
						id='toggle-datafilter-img'
					),
					dbc.Popover(
						'Ignore filters from the data view',
						target='btn-toggle-datafilter',
						body=True,
						trigger="hover",
						id='btn-toggle-datafilter-label',
					),
				],
				className='graph-overlay-item',
				id="btn-toggle-datafilter",
				n_clicks=0,
			),
		],
		className='graph-overlay-bl'
		),
		dbc.Offcanvas(
			[],
			id="node-info-canvas",
			title="",
			placement='end',
			backdrop=False,
			is_open=False,
			class_name='node-info-canvas',
			style={'borderLeft': 'none'}
		),
		# Search form
		html.Div([
			html.Img(src=get_asset_url('icons/graphpage/filter.svg')),
			html.Div(dcc.Input(
				value='',
				placeholder='🔎 IP or CIDR',
				type='text',
				pattern=r'^([1-9]|([1-9][0-9])|(1[0-9][0-9])|(2[0-4][0-9])|25[0-5])(\.([0-9]|([1-9][0-9])|(1[0-9][0-9])|(2[0-4][0-9])|25[0-5])){3}($|\/([1-9]|[1-2][0-9]|3[0-2])$)',
				id='subnet-filter-input',
				className='subnet-filter-input',
				debounce=True
			), id='filter-input-wrapper'),
			html.Div([], id='subnet-filter-badges', className='subnet-filter-badges'),
		], className='graph-search-menu'),
		html.Div([
			html.Div([
				html.Div([
					html.Span('0', id='ge-counter-edges', className='ge-counter-side'),
					html.Span(' edges', className='ge-counter-label ge-counter-side')
				]),
				html.Div([
					html.Span('{ ', className='ge-counter-label'),
					html.Span('0', id='ge-counter-eles'),
					html.Span(' }', className='ge-counter-label'),
#					html.Span(' elements', className='ge-counter-label')
				]),
				html.Div([
					html.Span('0', id='ge-counter-nodes', className='ge-counter-side'),
					html.Span(' nodes', className='ge-counter-label ge-counter-side')
				]),
			], className='ge-counter-flex')
		], className='graph-element-counter-container')
	],
	className='graph-container')


@callback(	Output('graph-settings', 'data', allow_duplicate=True),
		Output('subnet-filter-input', 'value'),
		Output('div-toaster', 'children', allow_duplicate=True),
		Input('subnet-filter-input', 'value'),
		State('graph-settings', 'data'),
		State('div-toaster', 'children'),
		prevent_initial_call=True
	)
def _cb_subnetFilterInput(value, graph_settings, toasts):
	if value is not None and value != '':
		if validateIP(value):
			if value in graph_settings['filters']['subnets']:
				toasts.append(CustomToast(
					[
						'Filter already applied.'
					],
					headerText='Filter invalid',
					level='warning'
				))
				return no_update, '', toasts

			graph_settings['filters']['subnets'].append(value)
			return graph_settings, '', no_update
		else:
			toasts.append(CustomToast(
				[
					'Invalid CIDR notation.'
				],
				headerText='Filter invalid',
				level='warning'
			))
			return no_update, no_update, toasts
	return no_update, no_update, no_update


@callback(	Output('graph-settings', 'data', allow_duplicate=True),
		Output('node-info-canvas', 'is_open', allow_duplicate=True),
		Input({'type': 'btn-del-filter', 'index': ALL}, 'n_clicks'),
		State('graph-settings', 'data'),
		prevent_initial_call=True
	)
def _cb_deleteInputFilter(btn, graph_settings):
	if any(btn):
		id = ctx.triggered_id['index']
		graph_settings['filters']['subnets'].pop(graph_settings['filters']['subnets'].index(id))
		return graph_settings, False
	return no_update, no_update


@callback(	Output('subnet-filter-badges', 'children'),
		Input('graph-settings', 'data')
	)
def _cb_updateSubnetBadges(graph_settings):
	return [
		dbc.Badge([
			html.Span(
				subnet,
				style={'marginRight': '0.5rem'},
				className = 'filter-badge-span'
			),
			html.Img(
				src=get_asset_url('icons/graphpage/trash.svg'),
				width=24,
				height=24,
				id={'type': 'btn-del-filter', 'index': subnet},
				className = 'filter-badge-trash-icon'
			)],
			pill=True,
			color='secondary',
			class_name='filter-badge'
		)
	for subnet in graph_settings['filters']['subnets']]


@callback(	Output('scan-graph', 'elements'),
		Output('div-toaster', 'children', allow_duplicate=True),
		Output('render-trigger', 'data'),
		Output('ge-counter-edges', 'children'),
		Output('ge-counter-eles', 'children'),
		Output('ge-counter-nodes', 'children'),
		Input('graph-settings', 'data'),
		State('div-toaster', 'children'),
		prevent_initial_call=True
	)
def _cb_redrawGraph(graph_settings, toasts):
	subnet_filter = graph_settings['filters']['subnets']

	# populate the input filters with the datable filters
	# this will be empty, unless the user wants to enable the datapage filters in the graph
	inp_filters = graph_settings['filters']['datatable']

	viewScans = graph_settings['scans-visible']
	viewPorts = graph_settings['ports-visible']
	viewHostsNP = graph_settings['hostsnp-visible'] # hosts with no ports

	# Retrieve hosts

	# if the user added subnet filters, these must be added as filters as well
	if len(subnet_filter) > 0:
		# we add an AND NEW GROUP filter (so a list) to make sure this condition is met for all results
		new_filter_group = []
		for ix in range(len(subnet_filter)):
			if not validateIP(subnet_filter[ix]):
				# This can never happen via GUI but just in case..
				continue
			new_filter_group.append({
				'op': 'AND' if ix==0 else 'OR', # only the group gets an "AND" the unique filters get an OR
				'field': 'host.ipv4',
				'type': 'in subnet' if '/' in subnet_filter [ix] else 'is', # either for subnet or single IP
				'input': subnet_filter[ix] # the query parser will do the processing for us
			})
		inp_filters.append(new_filter_group)

	if not viewHostsNP:
		# we add an AND NEW GROUP filter (same reason as for the subnet filter)
		inp_filters.append([{
			'op': 'AND',
			'field': 'host.portcount',
			'type': 'greater than',
			'input': '0'
		}])

	dfm = DataFilterMap()
	query, query_params = dfm.buildSQLQuery(inp_filters)

	filter_active = len(query_params) > 0

	eles = []
	edge_counter = 0
	node_counter = 0

	db_con, db = getDB()

	# Not using fetchall here should improve performance
	# otherwise all hosts would be copied into a list in memory before being processed
	qr = db.execute(query, query_params)
	relevant_hosts = []

	relevant_sids = []
	relevant_pids = []

	for host in qr:
		# since we use the big query, we don't need every column, only enough to create the node
		hid, os_name, os_family, hostLabel, _, _, _, _, _, _, _, h_sids, h_pids  = host

		relevant_hosts.append(hid)

		# add scans that weren't stored already to the relevant sids
		for h_sid in set(h_sids.split(',')):
			if h_sid not in relevant_sids:
				relevant_sids.append(int(h_sid))

		# a host may not have any open ports so h_pids may be None
		# add all ports relevant to this host to the relevant ports
		# ports are always unique to a host so we don't need to check if they were added already
		if h_pids:
			for h_pid in h_pids.split(','):
				relevant_pids.append(int(h_pid))

		os = getOS(os_family, os_name)

		eles.append({
			'data': {'id': f'host-{hid}', 'label': hostLabel},
			'classes': f'{os}host',
		})
		node_counter += 1

	if len(relevant_hosts) == 0:
		toasts.append(CustomToast(
			[
				'Try removing a filter.' if filter_active else 'Import a scan first.'
			],
			headerText='No hosts available',
			level='warning'
	        ))
		db_con.close()
		return eles, toasts, no_update, 0, 0, 0

	# Retrieve relevant scans (and edges)
	if viewScans:
		sid_list=[]
		qr = db.execute('SELECT s.sid, s.tool, sh.host FROM scans s INNER JOIN scans_hosts sh ON s.sid == sh.scan')
		for scan in qr:
			sid, tool, hid = scan

			if sid not in relevant_sids:
				continue

			# add scan node if it doesn't exist yet
			if sid not in sid_list:
				sid_list.append(sid)
				eles.append({
					'data': {'id': f'scan-{sid}', 'label': tool},
					'classes': 'scan'
				})
				node_counter += 1

			# add an edge for every host id
			# I believe filtering items in memory should be more efficient than a SQL query for every host
			if hid in relevant_hosts:
				eles.append({
					'data': {'source': f'scan-{sid}', 'target': f'host-{hid}', 'id': f'scan-edge-{sid}-{hid}', 'label': ''}
				})
				edge_counter += 1

	# Retrieve relevant ports
	if viewPorts:
		qr = db.execute('SELECT pid, port, protocol, host FROM ports')
		for port in qr:
			pid, portnumber, protocol, hid = port

			if pid not in relevant_pids:
				continue

			# Filter out irrelevant hosts
			if hid in relevant_hosts:
				eles.append({
					'data': {'id': f'port-{pid}', 'label': f'{portnumber}/{protocol}'},
					'classes': 'port'
				})
				node_counter += 1
				# Retrieve the relevant edges
				eles.append({
					'data': {'source': f'host-{hid}', 'target': f'port-{pid}', 'id': f'host-edge-{hid}-{pid}', 'label': ''}
				})
				edge_counter += 1

	db_con.close()

	ele_counter = node_counter + edge_counter

	if ele_counter > graph_settings['max_elements_render']:
		eles = []
		toasts.append(CustomToast(
			[
				f'You want to render too much data at once. The current limit is {graph_settings["max_elements_render"]} nodes & edges. Try a node filter (top left) or display filter (bottom left) to decrease the amount of nodes and edges.'
			],
			headerText=f'Layout warning ({ele_counter} elements)',
			level='warning',
			duration=15000
		))

	# return the counters regardless of whether we set the eles to []
	# this is because the user should know how many edges/nodes the query returned
	return eles, toasts, {'trigger': True}, edge_counter, ele_counter, node_counter


@callback(	Output('node-info-canvas', 'is_open'),
		Output('node-info-canvas', 'title'),
		Output('node-info-canvas', 'children'),
		Output('node-nav-state', 'data'),
		Output('scan-graph', 'elements', allow_duplicate=True),
		Input('scan-graph', 'tapNodeData'),
		Input({'type': 'btn-node-nav', 'index': ALL }, 'n_clicks'), # Input for dynamic nav (< & >) button that does not exist the first time
		Input({'type': 'btn-node-jump', 'index': ALL }, 'n_clicks'), # Input for dynamic jump buttons that does not exist the first time
		State('node-nav-state', 'data'),
		State('scan-graph', 'elements'),
		prevent_initial_call=True
	)
def _cb_selectNode(data, nav_clicks, jump_clicks, node_nav_state, node_elements):

	node_history = NodeHistory(node_nav_state)

	# When the nav menu was used instead of a click on a node
	_nav_prev = ctx.triggered_id=={'type': 'btn-node-nav', 'index': 'prev'} and nav_clicks[0]
	_nav_next = ctx.triggered_id=={'type': 'btn-node-nav', 'index': 'next'} and nav_clicks[1]
	_nav = _nav_prev or _nav_next

	_jumped = False
	if 'type' in ctx.triggered_id and 'index' in ctx.triggered_id and ctx.triggered_id['type'] == 'btn-node-jump':
		_jumped = True
		# If we did a jump from a position that was not the latest, add that position to latest again
		# so on moving back, we get to that node again - and not the current latest.
		if node_history.next_enabled():
			node_history.re_add_current()

	if _nav_prev:
		data = node_history.nav_previous()
	elif _nav_next:
		data = node_history.nav_next()
	elif _jumped:
		data = node_history.jump_to(ctx.triggered_id['index'])

	if data:
		db_con, db = getDB()

		patched_nodes = Patch()
		# Make sure only the active node is selected
		for index,element in enumerate(node_elements):
			patched_nodes[index]['selected'] = element['data']['id']==data['id']

		# If the node was clicked on in the graph or jumped to, then add it to the history
		if not _nav:
			node_history.add_node(data)

		node_type, node_id = data['id'].split('-') # the unused value is the unique identifier of each node
		node_label = data['label']

		# The label may update for hosts over time, so a history state may contain an outdated label value
		# That's why we need to update it when we display it coming from a nav
		# On clicks and jumps we can be sure that the label comes fresh from the database
		if node_type == 'host' and _nav:
			node_label = db.execute('SELECT label FROM hosts WHERE hid=?', (node_id,)).fetchone()[0]

		title_text = ''
		title_icon = ''
		match node_type:
			case 'host':
				title_text = f'{node_label}'
				title_icon = 'icons/graphpage/host.svg'
			case 'scan':
				title_text = f'{node_label} Scan'
				title_icon = 'icons/graphpage/terminal.svg'
			case 'port':
				title_text = f'Port {node_label}'
				title_icon = 'icons/graphpage/box.svg'
			case _:
				print(f'An unknown node type was selected: {data}')
				return False, no_update, no_update, no_update, no_update

		title = html.Div([
			html.Img(
				src=get_asset_url(title_icon),
				width=32,
				height=32,
				id='node-info-title-icon'
			),
			html.Span(
				title_text,
				style={'marginLeft': '0.5rem'}
			),
		])

		eles = []

		# Get node details from the database and build the node info panel

		if node_type == 'scan':
			sid = node_id
			sfid, stool, sargs, sversion, sstart, sstop, shosts, sup = db.execute('SELECT file, tool, args, version, datetime(start,"unixepoch"), datetime(stop,"unixepoch"), hostsScanned, hostsUp FROM scans WHERE sid=?', (sid,)).fetchone()

			sfilename = db.execute('SELECT filename FROM input_files WHERE fid=?', (sfid,)).fetchone()

			eles.append(dbc.ListGroup([
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/git-branch.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Tool version:',
							className='node-li-key'
						),
						html.Div(html.Span(
							f'{stool} {sversion}',
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/list.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Scan details:',
							className='node-li-key'
						),html.Div(html.Span(
							sargs,
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/file-text.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Filename:',
							className='node-li-key'
						),
						html.Div(html.Span(
							sfilename,
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/play-circle.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Scan started:',
							className='node-li-key'
						),
						html.Div(html.Span(
							f'{sstart}',
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/stop-circle.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Scan stopped:',
							className='node-li-key'
						),
						html.Div(html.Span(
							f'{sstop}',
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/pie-chart.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Hosts:',
							className='node-li-key'
						),
						html.Div(html.Span(
							f'{shosts} scanned, {sup} up',
							className='node-li-value',
						))
					]),
				],
				flush=True
			))
		if node_type == 'host':
			hid = node_id
			hipv4, hipv6, hmac, hreason, hos_name, hos_acc, hos_fam, hos_vendor, htag = db.execute('SELECT ipv4, ipv6, mac, reason, os_name, os_accuracy, os_family, os_vendor, tag FROM hosts WHERE hid=?', (hid,)).fetchone()
			hipv4 = hipv4 or -1
			hipv6 = hipv6 or ''
			hmac = hmac or ''
			hipv4 		= NumToIP(hipv4)
			hnames 		= [hn[0] for hn in db.execute('SELECT name FROM hostnames WHERE host=?', (hid,))]
			hos_acc 	= f' ({hos_acc:02}%)' if hos_acc>0 else ''
			hos_fam 	= f' {hos_fam}' if len(hos_fam) else ''
			hos_vendor 	= f' {hos_vendor}' if len(hos_vendor) else ''
			hos_name 	= hos_name+hos_fam+hos_vendor+hos_acc

			tools = db.execute('SELECT s.sid, tool FROM scans_hosts sh INNER JOIN scans s ON (sh.scan == s.sid) WHERE sh.host=?', (hid,)).fetchall()
			for tool in tools:
				sid, toolname = tool
				# generate a jump id like "Nmap-<sid>"
				node_history.add_jump(f'{toolname}-{sid}', {'id': f'scan-{sid}', 'label': f'{toolname}'})


			hports = db.execute('SELECT pid, port, protocol FROM ports WHERE host=?', (hid,)).fetchall()
			for hport in hports:
				# generate a jump id like "Port-<pid>"
				pid, port, protocol = hport
				node_history.add_jump(f'Port-{pid}', {'id': f'port-{pid}', 'label': f'{port}/{protocol}'})


			eles.append(dbc.ListGroup([
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/link.svg'),
							className='node-li-icon'
						),
						html.Span(
							'IPv4 address:',
							className='node-li-key'
						),
						html.Div(html.Span(
							hipv4,
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/link.svg'),
							className='node-li-icon'
						),
						html.Span(
							'IPv6 address:',
							className='node-li-key'
						),
						html.Div(html.Span(
							hipv6,
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/link.svg'),
							className='node-li-icon'
						),
						html.Span(
							'MAC address:',
							className='node-li-key'
						),
						html.Div(html.Span(
							hmac,
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/home.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Hostname:',
							className='node-li-key'
						),
						html.Div(html.Span(
							', '.join(hnames),
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/bookmark.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Tag:',
							className='node-li-key'
						),
						html.Div(html.Span(
							htag,
							className='node-li-value',
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/compass.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Up reason:',
							className='node-li-key'
						),
						html.Div(html.Span(
							hreason,
							className='node-li-value'
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/cpu.svg'),
							className='node-li-icon'
						),
						html.Span(
							'OS:',
							className='node-li-key'
						),
						html.Div(html.Span(
							hos_name,
							className='node-li-value'
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/graphpage/terminal.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Found by:',
							className='node-li-key'
						),
						html.Div([ dbc.Badge(
								f'{tool[1]} ({tool[0]})',
								pill=True,
								color='secondary',
								class_name='me-1 selectable-node-badge',
								id={'type': 'btn-node-jump', 'index': f'{tool[1]}-{tool[0]}'},
							) for tool in tools],
							className='node-li-value'
						)
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/graphpage/box.svg'),
							className='node-li-icon'
						),
						html.Span(
							f'Open ports ({len(hports)}):',
							className='node-li-key'
						),
						html.Div([ dbc.Badge(
								f'{hport[1]}/{hport[2]}',
								pill=True,
								color='secondary',
								class_name='me-1 selectable-node-badge',
								id={'type': 'btn-node-jump', 'index': f'Port-{hport[0]}'},
							) for hport in hports],
							className='node-li-value'
						)
					]),
				],
				flush=True
			))
		if node_type == 'port':
			pid = node_id
			hid, svc_name, svc_info, port, protocol = db.execute('SELECT host, svc_name, svc_info, port, protocol FROM ports WHERE pid=?', (pid,)).fetchone()
			hostlabel = db.execute('SELECT label FROM hosts WHERE hid=?', (hid,)).fetchone()[0]

			node_history.add_jump(f'Host-{hid}', {'id': f'host-{hid}', 'label': hostlabel}) # Jump ID for the host
			tools = db.execute('SELECT sid, tool FROM ports p LEFT JOIN scans_ports sp ON (p.pid=sp.port) LEFT JOIN scans s ON (sp.scan=s.sid) WHERE pid=?', (pid,)).fetchall()
			for tool in tools:
				sid, toolname = tool
				# generate a jump id like "Nmap-<sid>"
				node_history.add_jump(f'{toolname}-{sid}', {'id': f'scan-{sid}', 'label': f'{toolname}'})

			n_hosts = db.execute('SELECT COUNT(host) FROM ports WHERE port=? AND protocol=? AND host!=?', (port, protocol,hid)).fetchone()[0]

			qr = db.execute('SELECT name FROM pscripts WHERE host=? AND port=?', (hid,pid)).fetchall()
			script_names = ', '.join([name[0] for name in qr])

			eles.append(dbc.ListGroup([
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/graphpage/host.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Host:',
							className='node-li-key'
						),
						html.Div(html.Span(
							hostlabel,
							className='node-li-value span-link',
							id={'type': 'btn-node-jump', 'index': f'Host-{hid}'},
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/tool.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Service:',
							className='node-li-key'
						),
						html.Div(html.Span(
							svc_name,
							className='node-li-value'
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/activity.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Service info:',
							className='node-li-key'
						),
						html.Div(html.Span(
							svc_info,
							className='node-li-value'
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/trello.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Available script results:',
							className='node-li-key'
						),
						html.Div(html.Span(
							script_names,
							className='node-li-value'
						))
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/graphpage/terminal.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Found by:',
							className='node-li-key'
						),
						html.Div([ dbc.Badge(
								f'{tool[1]} ({tool[0]})',
								pill=True,
								color='secondary',
								class_name='me-1 selectable-node-badge',
								id={'type': 'btn-node-jump', 'index': f'{tool[1]}-{tool[0]}'},
							) for tool in tools],
							className='node-li-value'
						)
					]),
					dbc.ListGroupItem([
						html.Img(
							src=get_asset_url('icons/node-info/repeat.svg'),
							className='node-li-icon'
						),
						html.Span(
							'Also found on:',
							className='node-li-key'
						),
						html.Div(html.Span(
							f'{n_hosts} other host{"s" if n_hosts!=1 else ""}',
							className='node-li-value'
						))
					]),
				],
				flush=True
			))

		eles.append(html.Hr())
		eles.append(html.Div([
			dbc.Button([
					html.Img(
						src=get_asset_url('icons/node-info/chevron-left.svg'),
					),
				],
				disabled=not node_history.prev_enabled(),
				color='dark',
				class_name='node-info-nav-btn',
				id={'type': 'btn-node-nav', 'index': 'prev'},
				n_clicks=0,
			),
			dbc.Button([
					html.Img(
						src=get_asset_url('icons/node-info/chevron-right.svg'),
					),
				],
				disabled=not node_history.next_enabled(),
				color='dark',
				class_name='node-info-nav-btn',
				id={'type': 'btn-node-nav', 'index': 'next'},
				n_clicks=0,
			)
		], className='node-info-nav-container'))

		db_con.close()

		return True, title, eles, node_history.get(), patched_nodes
	return False, no_update, no_update, no_update, no_update
